# Copyright (c) 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty
import warnings
# TODO(mordred) This is only here to ease the OSC transition

from openstack.config import cloud_region
from openstack import warnings as os_warnings

if ty.TYPE_CHECKING:
    from keystoneauth1 import discover
    from keystoneauth1 import plugin
    from keystoneauth1 import session as ks_session
    import prometheus_client

    from openstack.config import loader


class CloudConfig(cloud_region.CloudRegion):
    def __init__(
        self,
        name: str | None,
        region: str | None,
        config: dict[str, ty.Any] | None,
        force_ipv4: bool = False,
        auth_plugin: ty.Optional['plugin.BaseAuthPlugin'] = None,
        openstack_config: ty.Optional['loader.OpenStackConfig'] = None,
        session_constructor: type['ks_session.Session'] | None = None,
        app_name: str | None = None,
        app_version: str | None = None,
        session: ty.Optional['ks_session.Session'] = None,
        discovery_cache: dict[str, 'discover.Discover'] | None = None,
        extra_config: dict[str, ty.Any] | None = None,
        cache_expiration_time: int = 0,
        cache_expirations: dict[str, int] | None = None,
        cache_path: str | None = None,
        cache_class: str = 'dogpile.cache.null',
        cache_arguments: dict[str, ty.Any] | None = None,
        password_callback: cloud_region._PasswordCallback | None = None,
        statsd_host: str | None = None,
        statsd_port: str | None = None,
        statsd_prefix: str | None = None,
        # TODO(stephenfin): Add better types
        influxdb_config: dict[str, ty.Any] | None = None,
        collector_registry: ty.Optional[
            'prometheus_client.CollectorRegistry'
        ] = None,
        cache_auth: bool = False,
    ) -> None:
        warnings.warn(
            'The CloudConfig class has been deprecated in favour of '
            'CloudRegion. Please update your references.',
            os_warnings.RemovedInSDK60Warning,
        )

        self.region = region

        super().__init__(
            name,
            region,
            config,
            force_ipv4=force_ipv4,
            auth_plugin=auth_plugin,
            openstack_config=openstack_config,
            session_constructor=session_constructor,
            app_name=app_name,
            app_version=app_version,
            session=session,
            discovery_cache=discovery_cache,
            extra_config=extra_config,
            cache_expiration_time=cache_expiration_time,
            cache_expirations=cache_expirations,
            cache_path=cache_path,
            cache_class=cache_class,
            cache_arguments=cache_arguments,
            password_callback=password_callback,
            statsd_host=statsd_host,
            statsd_port=statsd_port,
            statsd_prefix=statsd_prefix,
            influxdb_config=influxdb_config,
            collector_registry=collector_registry,
            cache_auth=cache_auth,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.unit import test_proxy_base
from openstack.workflow.v2 import _proxy
from openstack.workflow.v2 import cron_trigger
from openstack.workflow.v2 import execution
from openstack.workflow.v2 import workflow


class TestWorkflowProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_workflows(self):
        self.verify_list(self.proxy.workflows, workflow.Workflow)

    def test_executions(self):
        self.verify_list(self.proxy.executions, execution.Execution)

    def test_workflow_get(self):
        self.verify_get(self.proxy.get_workflow, workflow.Workflow)

    def test_execution_get(self):
        self.verify_get(self.proxy.get_execution, execution.Execution)

    def test_workflow_create(self):
        self.verify_create(self.proxy.create_workflow, workflow.Workflow)

    def test_workflow_update(self):
        self.verify_update(self.proxy.update_workflow, workflow.Workflow)

    def test_execution_create(self):
        self.verify_create(self.proxy.create_execution, execution.Execution)

    def test_workflow_delete(self):
        self.verify_delete(self.proxy.delete_workflow, workflow.Workflow, True)

    def test_execution_delete(self):
        self.verify_delete(
            self.proxy.delete_execution, execution.Execution, True
        )

    def test_workflow_find(self):
        self.verify_find(self.proxy.find_workflow, workflow.Workflow)

    def test_execution_find(self):
        self.verify_find(self.proxy.find_execution, execution.Execution)


class TestCronTriggerProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_cron_triggers(self):
        self.verify_list(self.proxy.cron_triggers, cron_trigger.CronTrigger)

    def test_cron_trigger_get(self):
        self.verify_get(self.proxy.get_cron_trigger, cron_trigger.CronTrigger)

    def test_cron_trigger_create(self):
        self.verify_create(
            self.proxy.create_cron_trigger, cron_trigger.CronTrigger
        )

    def test_cron_trigger_delete(self):
        self.verify_delete(
            self.proxy.delete_cron_trigger, cron_trigger.CronTrigger, True
        )

    def test_cron_trigger_find(self):
        self.verify_find(
            self.proxy.find_cron_trigger,
            cron_trigger.CronTrigger,
            expected_kwargs={'all_projects': False},
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from openstack import exceptions
from openstack import resource
from openstack import utils


class Attachment(resource.Resource):
    resource_key = "attachment"
    resources_key = "attachments"
    base_path = "/attachments"

    # capabilities
    allow_create = True
    allow_delete = True
    allow_commit = True
    allow_list = True
    allow_fetch = True

    _max_microversion = "3.54"

    # Properties
    #: The ID of the attachment.
    id = resource.Body("id")
    #: The status of the attachment.
    status = resource.Body("status")
    #: The UUID of the attaching instance.
    instance = resource.Body("instance")
    #: The UUID of the volume which the attachment belongs to.
    volume_id = resource.Body("volume_id")
    #: The time when attachment is attached.
    attached_at = resource.Body("attach_time")
    #: The time when attachment is detached.
    detached_at = resource.Body("detach_time")
    #: The attach mode of attachment, read-only ('ro') or read-and-write
    # ('rw'), default is 'rw'.
    attach_mode = resource.Body("mode")
    #: The connection info used for server to connect the volume.
    connection_info = resource.Body("connection_info")
    #: The connector object.
    connector = resource.Body("connector")

    def create(
        self,
        session,
        prepend_key=True,
        base_path=None,
        *,
        resource_request_key=None,
        resource_response_key=None,
        microversion=None,
        **params,
    ):
        if utils.supports_microversion(session, '3.54'):
            if not self.attach_mode:
                self._body.clean(only={'mode'})
        return super().create(
            session,
            prepend_key=prepend_key,
            base_path=base_path,
            resource_request_key=resource_request_key,
            resource_response_key=resource_response_key,
            microversion=microversion,
            **params,
        )

    def complete(self, session, *, microversion=None):
        """Mark the attachment as completed."""
        body = {'os-complete': self.id}
        if not microversion:
            microversion = self._get_microversion(session)
        url = os.path.join(Attachment.base_path, self.id, 'action')
        response = session.post(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)

    def _prepare_request_body(
        self,
        patch,
        prepend_key,
        *,
        resource_request_key=None,
    ):
        body = self._body.dirty
        if body.get('volume_id'):
            body['volume_uuid'] = body.pop('volume_id')
        if body.get('instance'):
            body['instance_uuid'] = body.pop('instance')
        if prepend_key and self.resource_key is not None:
            body = {self.resource_key: body}
        return body



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.network.v2 import agent
from openstack.tests.functional import base


class TestAgent(base.BaseFunctionalTest):
    AGENT: agent.Agent
    DESC = "test description"

    def validate_uuid(self, s):
        try:
            uuid.UUID(s)
        except Exception:
            return False
        return True

    def setUp(self):
        super().setUp()
        if not self.user_cloud._has_neutron_extension("agent"):
            self.skipTest("Neutron agent extension is required for this test")

        agent_list = list(self.user_cloud.network.agents())
        if len(agent_list) == 0:
            self.skipTest("No agents available")
        self.AGENT = agent_list[0]
        assert isinstance(self.AGENT, agent.Agent)

    def test_list(self):
        agent_list = list(self.user_cloud.network.agents())
        self.AGENT = agent_list[0]
        assert isinstance(self.AGENT, agent.Agent)
        self.assertTrue(self.validate_uuid(self.AGENT.id))

    def test_get(self):
        sot = self.user_cloud.network.get_agent(self.AGENT.id)
        self.assertEqual(self.AGENT.id, sot.id)

    def test_update(self):
        sot = self.user_cloud.network.update_agent(
            self.AGENT.id, description=self.DESC
        )
        self.assertEqual(self.DESC, sot.description)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import bgp_peer
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'auth_type': 'none',
    'remote_as': '1001',
    'name': 'bgp-peer',
    'peer_ip': '10.0.0.3',
    'id': IDENTIFIER,
    'project_id': '42',
}


class TestBgpPeer(base.TestCase):
    def test_basic(self):
        sot = bgp_peer.BgpPeer()
        self.assertEqual('bgp_peer', sot.resource_key)
        self.assertEqual('bgp_peers', sot.resources_key)
        self.assertEqual('/bgp-peers', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = bgp_peer.BgpPeer(**EXAMPLE)
        self.assertEqual(EXAMPLE['auth_type'], sot.auth_type)
        self.assertEqual(EXAMPLE['remote_as'], sot.remote_as)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['peer_ip'], sot.peer_ip)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)

        self.assertDictEqual(
            {
                'limit': 'limit',
                'marker': 'marker',
            },
            sot._query_mapping._mapping,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import share_snapshot
from openstack.tests.unit import base


EXAMPLE = {
    "status": "creating",
    "share_id": "406ea93b-32e9-4907-a117-148b3945749f",
    "user_id": "5c7bdb6eb0504d54a619acf8375c08ce",
    "name": "snapshot_share1",
    "created_at": "2021-06-07T11:50:39.756808",
    "description": "Here is a snapshot of share Share1",
    "share_proto": "NFS",
    "share_size": 1,
    "id": "6d221c1d-0200-461e-8d20-24b4776b9ddb",
    "project_id": "cadd7139bc3148b8973df097c0911016",
    "size": 1,
}


class TestShareSnapshot(base.TestCase):
    def test_basic(self):
        snapshot_resource = share_snapshot.ShareSnapshot()
        self.assertEqual('snapshots', snapshot_resource.resources_key)
        self.assertEqual('/snapshots', snapshot_resource.base_path)
        self.assertTrue(snapshot_resource.allow_list)

        self.assertDictEqual(
            {
                "limit": "limit",
                "marker": "marker",
                "snapshot_id": "snapshot_id",
            },
            snapshot_resource._query_mapping._mapping,
        )

    def test_make_share_snapshot(self):
        snapshot_resource = share_snapshot.ShareSnapshot(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], snapshot_resource.id)
        self.assertEqual(EXAMPLE['share_id'], snapshot_resource.share_id)
        self.assertEqual(EXAMPLE['user_id'], snapshot_resource.user_id)
        self.assertEqual(EXAMPLE['created_at'], snapshot_resource.created_at)
        self.assertEqual(EXAMPLE['status'], snapshot_resource.status)
        self.assertEqual(EXAMPLE['name'], snapshot_resource.name)
        self.assertEqual(EXAMPLE['description'], snapshot_resource.description)
        self.assertEqual(EXAMPLE['share_proto'], snapshot_resource.share_proto)
        self.assertEqual(EXAMPLE['share_size'], snapshot_resource.share_size)
        self.assertEqual(EXAMPLE['project_id'], snapshot_resource.project_id)
        self.assertEqual(EXAMPLE['size'], snapshot_resource.size)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import port_forwarding
from openstack.tests.unit import base


EXAMPLE = {
    'id': 'pf_id',
    'protocol': 'tcp',
    'internal_ip_address': '1.2.3.4',
    'floatingip_id': 'floating-ip-uuid',
    'internal_port': 80,
    'internal_port_id': 'internal-port-uuid',
    'external_port': 8080,
    'description': 'description',
}


class TestFloatingIP(base.TestCase):
    def test_basic(self):
        sot = port_forwarding.PortForwarding()
        self.assertEqual('port_forwarding', sot.resource_key)
        self.assertEqual('port_forwardings', sot.resources_key)
        self.assertEqual(
            '/floatingips/%(floatingip_id)s/port_forwardings', sot.base_path
        )
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                'internal_port_id': 'internal_port_id',
                'external_port': 'external_port',
                'limit': 'limit',
                'marker': 'marker',
                'protocol': 'protocol',
                'sort_dir': 'sort_dir',
                'sort_key': 'sort_key',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = port_forwarding.PortForwarding(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['floatingip_id'], sot.floatingip_id)
        self.assertEqual(EXAMPLE['protocol'], sot.protocol)
        self.assertEqual(
            EXAMPLE['internal_ip_address'], sot.internal_ip_address
        )
        self.assertEqual(EXAMPLE['internal_port'], sot.internal_port)
        self.assertEqual(EXAMPLE['internal_port_id'], sot.internal_port_id)
        self.assertEqual(EXAMPLE['external_port'], sot.external_port)
        self.assertEqual(EXAMPLE['description'], sot.description)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.unit import base
from openstack.workflow.v2 import workflow


FAKE = {
    'scope': 'private',
    'id': 'ffaed25e-46f5-4089-8e20-b3b4722fd597',
    'definition': 'workflow_def',
}


class TestWorkflow(base.TestCase):
    def setUp(self):
        super().setUp()

    def test_basic(self):
        sot = workflow.Workflow()
        self.assertEqual('workflow', sot.resource_key)
        self.assertEqual('workflows', sot.resources_key)
        self.assertEqual('/workflows', sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_list)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)

    def test_instantiate(self):
        sot = workflow.Workflow(**FAKE)
        self.assertEqual(FAKE['id'], sot.id)
        self.assertEqual(FAKE['scope'], sot.scope)
        self.assertEqual(FAKE['definition'], sot.definition)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.key_manager.v1 import secret_store
from openstack.tests.unit import base


EXAMPLE = {
    "status": "ACTIVE",
    "updated": "2016-08-22T23:46:45.114283",
    "name": "PKCS11 HSM",
    "created": "2016-08-22T23:46:45.114283",
    "secret_store_ref": "http://localhost:9311/v1/secret-stores/4d27b7a7-b82f-491d-88c0-746bd67dadc8",
    "global_default": True,
    "crypto_plugin": "p11_crypto",
    "secret_store_plugin": "store_crypto",
}


class TestSecretStore(base.TestCase):
    def test_basic(self):
        sot = secret_store.SecretStore()
        self.assertEqual('secret_stores', sot.resources_key)
        self.assertEqual('/secret-stores', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = secret_store.SecretStore(**EXAMPLE)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['created'], sot.created_at)
        self.assertEqual(EXAMPLE['updated'], sot.updated_at)
        self.assertEqual(EXAMPLE['secret_store_ref'], sot.secret_store_ref)
        self.assertEqual(EXAMPLE['global_default'], sot.global_default)
        self.assertEqual(EXAMPLE['crypto_plugin'], sot.crypto_plugin)
        self.assertEqual(
            EXAMPLE['secret_store_plugin'], sot.secret_store_plugin
        )
        # Test the alternate_id extraction
        self.assertEqual(
            '4d27b7a7-b82f-491d-88c0-746bd67dadc8', sot.secret_store_id
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from openstack.clustering.v1 import cluster
from openstack.tests.unit import base


CLUSTERING_DICT = {
    'name': 'fake-name',
    'profile_id': '1',
    'desired_capacity': 1,
    'config': {'a': 'b'},
    'max_size': 1,
    'min_size': 1,
    'timeout': 100,
    'metadata': {},
}

PROFILE_DICT = {'name': 'fake-profile-name', 'spec': {}, 'metadata': {}}

POLICY_DICT = {
    'name': 'fake-profile-name',
    'spec': {},
}

RECEIVER_DICT = {
    'action': 'FAKE_CLUSTER_ACTION',
    'cluster_id': 'fake-cluster-id',
    'name': 'fake-receiver-name',
    'params': {},
    'type': 'webhook',
}

NEW_CLUSTERING_DICT = copy.copy(CLUSTERING_DICT)
NEW_CLUSTERING_DICT['id'] = '1'
NEW_PROFILE_DICT = copy.copy(PROFILE_DICT)
NEW_PROFILE_DICT['id'] = '1'
NEW_POLICY_DICT = copy.copy(POLICY_DICT)
NEW_POLICY_DICT['id'] = '1'
NEW_RECEIVER_DICT = copy.copy(RECEIVER_DICT)
NEW_RECEIVER_DICT['id'] = '1'


class TestClustering(base.TestCase):
    def assertAreInstances(self, elements, elem_type):
        for e in elements:
            self.assertIsInstance(e, elem_type)

    def _compare_clusters(self, exp, real):
        self.assertEqual(
            cluster.Cluster(**exp).to_dict(computed=False),
            real.to_dict(computed=False),
        )

    def setUp(self):
        super().setUp()
        self.use_senlin()


#    def test_create_cluster(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles', '1']),
#                 json={
#                     "profiles": [NEW_PROFILE_DICT]}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles']),
#                 json={
#                     "profiles": [NEW_PROFILE_DICT]}),
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters']),
#                 json=NEW_CLUSTERING_DICT)
#        ])
#        profile = self.cloud.get_cluster_profile_by_id(NEW_PROFILE_DICT['id'])
#        c = self.cloud.create_cluster(
#            name=CLUSTERING_DICT['name'],
#            desired_capacity=CLUSTERING_DICT['desired_capacity'],
#            profile=profile,
#            config=CLUSTERING_DICT['config'],
#            max_size=CLUSTERING_DICT['max_size'],
#            min_size=CLUSTERING_DICT['min_size'],
#            metadata=CLUSTERING_DICT['metadata'],
#            timeout=CLUSTERING_DICT['timeout'])
#
#        self._compare_clusters(NEW_CLUSTERING_DICT, c)
#        self.assert_calls()
#
#    def test_create_cluster_exception(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles', '1']),
#                 json={
#                     "profiles": [NEW_PROFILE_DICT]}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles']),
#                 json={
#                     "profiles": [NEW_PROFILE_DICT]}),
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters']),
#                 status_code=500)
#        ])
#        profile = self.cloud.get_cluster_profile_by_id(NEW_PROFILE_DICT['id'])
#        with testtools.ExpectedException(
#                exc.OpenStackCloudHTTPError):
#            self.cloud.create_cluster(name='fake-name', profile=profile)
#        self.assert_calls()
#
#    def test_get_cluster_by_id(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1']),
#                 json={
#                     "cluster": NEW_CLUSTERING_DICT})
#        ])
#        cluster = self.cloud.get_cluster_by_id('1')
#        self.assertEqual(cluster['id'], '1')
#        self.assert_calls()
#
#    def test_get_cluster_not_found_returns_false(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters',
#                                                     'no-cluster']),
#                 status_code=404)
#        ])
#        c = self.cloud.get_cluster_by_id('no-cluster')
#        self.assertFalse(c)
#        self.assert_calls()
#
#    def test_update_cluster(self):
#        new_max_size = 5
#        updated_cluster = copy.copy(NEW_CLUSTERING_DICT)
#        updated_cluster['max_size'] = new_max_size
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1']),
#                 json={
#                     "cluster": NEW_CLUSTERING_DICT}),
#            dict(method='PATCH',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1']),
#                 json=updated_cluster,
#                 )
#        ])
#        cluster = self.cloud.get_cluster_by_id('1')
#        c = self.cloud.update_cluster(cluster, new_max_size)
#        self.assertEqual(updated_cluster, c)
#        self.assert_calls()
#
#    def test_delete_cluster(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters']),
#                 json={
#                     "clusters": [NEW_CLUSTERING_DICT]}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1',
#                                                     'policies']),
#                 json={"cluster_policies": []}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers']),
#                 json={"receivers": []}),
#            dict(method='DELETE',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1']),
#                 json=NEW_CLUSTERING_DICT)
#        ])
#        self.assertTrue(self.cloud.delete_cluster('1'))
#        self.assert_calls()
#
#    def test_list_clusters(self):
#        clusters = {'clusters': [NEW_CLUSTERING_DICT]}
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters']),
#                 json=clusters)
#        ])
#        c = self.cloud.list_clusters()
#
#        self.assertIsInstance(c, list)
#        self.assertAreInstances(c, dict)
#
#        self.assert_calls()
#
#    def test_attach_policy_to_cluster(self):
#        policy = {
#            'policy_id': '1',
#            'enabled': 'true'
#        }
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1']),
#                 json={
#                     "cluster": NEW_CLUSTERING_DICT}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies', '1']),
#                 json={
#                     "policy": NEW_POLICY_DICT}),
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1',
#                                                     'actions']),
#                 json={'policy_attach': policy})
#        ])
#        cluster = self.cloud.get_cluster_by_id('1')
#        policy = self.cloud.get_cluster_policy_by_id('1')
#        p = self.cloud.attach_policy_to_cluster(cluster, policy, 'true')
#        self.assertTrue(p)
#        self.assert_calls()
#
#    def test_detach_policy_from_cluster(self):
#        updated_cluster = copy.copy(NEW_CLUSTERING_DICT)
#        updated_cluster['policies'] = ['1']
#        detached_cluster = copy.copy(NEW_CLUSTERING_DICT)
#        detached_cluster['policies'] = []
#
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1']),
#                 json={
#                     "cluster": NEW_CLUSTERING_DICT}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies', '1']),
#                 json={
#                     "policy": NEW_POLICY_DICT}),
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1',
#                                                     'actions']),
#                 json={'policy_detach': {'policy_id': '1'}}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1']),
#                 json={
#                     "cluster": updated_cluster}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1']),
#                 json={
#                     "cluster": detached_cluster}),
#        ])
#        cluster = self.cloud.get_cluster_by_id('1')
#        policy = self.cloud.get_cluster_policy_by_id('1')
#        p = self.cloud.detach_policy_from_cluster(cluster, policy, wait=True)
#        self.assertTrue(p)
#        self.assert_calls()
#
#    def test_get_policy_on_cluster_by_id(self):
#        cluster_policy = {
#            "cluster_id": "1",
#            "cluster_name": "cluster1",
#            "enabled": True,
#            "id": "1",
#            "policy_id": "1",
#            "policy_name": "policy1",
#            "policy_type": "senlin.policy.deletion-1.0"
#        }
#
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1',
#                                                     'policies', '1']),
#                 json={
#                     "cluster_policy": cluster_policy})
#        ])
#        policy = self.cloud.get_policy_on_cluster('1', '1')
#        self.assertEqual(policy['cluster_id'], '1')
#        self.assert_calls()
#
#    def test_get_policy_on_cluster_not_found_returns_false(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1',
#                                                     'policies',
#              'no-policy']),
#                 status_code=404)
#        ])
#        p = self.cloud.get_policy_on_cluster('1', 'no-policy')
#        self.assertFalse(p)
#        self.assert_calls()
#
#    def test_update_policy_on_cluster(self):
#        policy = {
#            'policy_id': '1',
#            'enabled': 'true'
#        }
#        updated_cluster = copy.copy(NEW_CLUSTERING_DICT)
#        updated_cluster['policies'] = policy
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1']),
#                 json={
#                     "cluster": NEW_CLUSTERING_DICT}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies',
#                                                     '1']),
#                 json={
#                     "policy": NEW_POLICY_DICT}),
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1',
#                                                     'actions']),
#                 json={'policies': []})
#        ])
#        cluster = self.cloud.get_cluster_by_id('1')
#        policy = self.cloud.get_cluster_policy_by_id('1')
#        p = self.cloud.update_policy_on_cluster(cluster, policy, True)
#        self.assertTrue(p)
#        self.assert_calls()
#
#    def test_get_policy_on_cluster(self):
#        cluster_policy = {
#            'cluster_id': '1',
#            'cluster_name': 'cluster1',
#            'enabled': 'true',
#            'id': '1',
#            'policy_id': '1',
#            'policy_name': 'policy1',
#            'policy_type': 'type'
#        }
#
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters', '1',
#                                                     'policies', '1']),
#                 json={
#                     "cluster_policy": cluster_policy})
#        ])
#        get_policy = self.cloud.get_policy_on_cluster('1', '1')
#        self.assertEqual(get_policy, cluster_policy)
#        self.assert_calls()
#
#    def test_create_cluster_profile(self):
#        self.register_uris([
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles']),
#                 json={'profile': NEW_PROFILE_DICT})
#        ])
#        p = self.cloud.create_cluster_profile('fake-profile-name', {})
#
#        self.assertEqual(NEW_PROFILE_DICT, p)
#        self.assert_calls()
#
#    def test_create_cluster_profile_exception(self):
#        self.register_uris([
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles']),
#                 status_code=500)
#        ])
#        with testtools.ExpectedException(
#                exc.OpenStackCloudHTTPError,
#                "Error creating profile fake-profile-name.*"):
#            self.cloud.create_cluster_profile('fake-profile-name', {})
#        self.assert_calls()
#
#    def test_list_cluster_profiles(self):
#        profiles = {'profiles': [NEW_PROFILE_DICT]}
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles']),
#                 json=profiles)
#        ])
#        p = self.cloud.list_cluster_profiles()
#
#        self.assertIsInstance(p, list)
#        self.assertAreInstances(p, dict)
#
#        self.assert_calls()
#
#    def test_get_cluster_profile_by_id(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles', '1']),
#                 json={
#                     "profile": NEW_PROFILE_DICT})
#        ])
#        p = self.cloud.get_cluster_profile_by_id('1')
#        self.assertEqual(p['id'], '1')
#        self.assert_calls()
#
#    def test_get_cluster_profile_not_found_returns_false(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles',
#                                                     'no-profile']),
#                 status_code=404)
#        ])
#        p = self.cloud.get_cluster_profile_by_id('no-profile')
#        self.assertFalse(p)
#        self.assert_calls()
#
#    def test_update_cluster_profile(self):
#        new_name = "new-name"
#        updated_profile = copy.copy(NEW_PROFILE_DICT)
#        updated_profile['name'] = new_name
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles']),
#                 json={
#                     "profiles": [NEW_PROFILE_DICT]}),
#            dict(method='PATCH',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles', '1']),
#                 json=updated_profile,
#                 )
#        ])
#        p = self.cloud.update_cluster_profile('1', new_name=new_name)
#        self.assertEqual(updated_profile, p)
#        self.assert_calls()
#
#    def test_delete_cluster_profile(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles', '1']),
#                 json={
#                     "profile": NEW_PROFILE_DICT}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters']),
#                 json={'clusters': [{'cluster': CLUSTERING_DICT}]}),
#            dict(method='DELETE',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'profiles', '1']),
#                 json=NEW_PROFILE_DICT)
#        ])
#        profile = self.cloud.get_cluster_profile_by_id('1')
#        self.assertTrue(self.cloud.delete_cluster_profile(profile))
#        self.assert_calls()
#
#    def test_create_cluster_policy(self):
#        self.register_uris([
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies']),
#                 json={'policy': NEW_POLICY_DICT})
#        ])
#        p = self.cloud.create_cluster_policy('fake-policy-name', {})
#
#        self.assertEqual(NEW_POLICY_DICT, p)
#        self.assert_calls()
#
#    def test_create_cluster_policy_exception(self):
#        self.register_uris([
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies']),
#                 status_code=500)
#        ])
#        with testtools.ExpectedException(
#                exc.OpenStackCloudHTTPError,
#                "Error creating policy fake-policy-name.*"):
#            self.cloud.create_cluster_policy('fake-policy-name', {})
#        self.assert_calls()
#
#    def test_list_cluster_policies(self):
#        policies = {'policies': [NEW_POLICY_DICT]}
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies']),
#                 json=policies)
#        ])
#        p = self.cloud.list_cluster_policies()
#
#        self.assertIsInstance(p, list)
#        self.assertAreInstances(p, dict)
#
#        self.assert_calls()
#
#    def test_get_cluster_policy_by_id(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies', '1']),
#                 json={
#                     "policy": NEW_POLICY_DICT})
#        ])
#        p = self.cloud.get_cluster_policy_by_id('1')
#        self.assertEqual(p['id'], '1')
#        self.assert_calls()
#
#    def test_get_cluster_policy_not_found_returns_false(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies',
#                                                     'no-policy']),
#                 status_code=404)
#        ])
#        p = self.cloud.get_cluster_policy_by_id('no-policy')
#        self.assertFalse(p)
#        self.assert_calls()
#
#    def test_update_cluster_policy(self):
#        new_name = "new-name"
#        updated_policy = copy.copy(NEW_POLICY_DICT)
#        updated_policy['name'] = new_name
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies']),
#                 json={
#                     "policies": [NEW_POLICY_DICT]}),
#            dict(method='PATCH',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies', '1']),
#                 json=updated_policy,
#                 )
#        ])
#        p = self.cloud.update_cluster_policy('1', new_name=new_name)
#        self.assertEqual(updated_policy, p)
#        self.assert_calls()
#
#    def test_delete_cluster_policy(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies', '1']),
#                 json={
#                     "policy": NEW_POLICY_DICT}),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters']),
#                 json={}),
#            dict(method='DELETE',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'policies', '1']),
#                 json=NEW_POLICY_DICT)
#        ])
#        self.assertTrue(self.cloud.delete_cluster_policy('1'))
#        self.assert_calls()
#
#    def test_create_cluster_receiver(self):
#        clusters = {'clusters': [NEW_CLUSTERING_DICT]}
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters']),
#                 json=clusters),
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers']),
#                 json={'receiver': NEW_RECEIVER_DICT})
#        ])
#        r = self.cloud.create_cluster_receiver('fake-receiver-name', {})
#
#        self.assertEqual(NEW_RECEIVER_DICT, r)
#        self.assert_calls()
#
#    def test_create_cluster_receiver_exception(self):
#        clusters = {'clusters': [NEW_CLUSTERING_DICT]}
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'clusters']),
#                 json=clusters),
#            dict(method='POST',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers']),
#                 status_code=500),
#        ])
#        with testtools.ExpectedException(
#                exc.OpenStackCloudHTTPError,
#                "Error creating receiver fake-receiver-name.*"):
#            self.cloud.create_cluster_receiver('fake-receiver-name', {})
#        self.assert_calls()
#
#    def test_list_cluster_receivers(self):
#        receivers = {'receivers': [NEW_RECEIVER_DICT]}
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers']),
#                 json=receivers)
#        ])
#        r = self.cloud.list_cluster_receivers()
#
#        self.assertIsInstance(r, list)
#        self.assertAreInstances(r, dict)
#
#        self.assert_calls()
#
#    def test_get_cluster_receiver_by_id(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers', '1']),
#                 json={
#                     "receiver": NEW_RECEIVER_DICT})
#        ])
#        r = self.cloud.get_cluster_receiver_by_id('1')
#        self.assertEqual(r['id'], '1')
#        self.assert_calls()
#
#    def test_get_cluster_receiver_not_found_returns_false(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers',
#                                                     'no-receiver']),
#                 json={'receivers': []})
#        ])
#        p = self.cloud.get_cluster_receiver_by_id('no-receiver')
#        self.assertFalse(p)
#        self.assert_calls()
#
#    def test_update_cluster_receiver(self):
#        new_name = "new-name"
#        updated_receiver = copy.copy(NEW_RECEIVER_DICT)
#        updated_receiver['name'] = new_name
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers']),
#                 json={
#                     "receivers": [NEW_RECEIVER_DICT]}),
#            dict(method='PATCH',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers', '1']),
#                 json=updated_receiver,
#                 )
#        ])
#        r = self.cloud.update_cluster_receiver('1', new_name=new_name)
#        self.assertEqual(updated_receiver, r)
#        self.assert_calls()
#
#    def test_delete_cluster_receiver(self):
#        self.register_uris([
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers']),
#                 json={
#                     "receivers": [NEW_RECEIVER_DICT]}),
#            dict(method='DELETE',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers', '1']),
#                 json=NEW_RECEIVER_DICT),
#            dict(method='GET',
#                 uri=self.get_mock_url(
#                     'clustering', 'public', append=['v1', 'receivers', '1']),
#                 json={}),
#        ])
#        self.assertTrue(self.cloud.delete_cluster_receiver('1', wait=True))
#        self.assert_calls()



# Copyright 2017 OVH SAS
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import testtools

from openstack import exceptions
from openstack.network.v2 import port as _port
from openstack.network.v2 import router as _router
from openstack.tests.unit import base


class TestRouter(base.TestCase):
    router_name = 'goofy'
    router_id = '57076620-dcfb-42ed-8ad6-79ccb4a79ed2'
    subnet_id = '1f1696eb-7f47-47f6-835c-4889bff88604'

    mock_router_rep = {
        'admin_state_up': True,
        'availability_zone_hints': [],
        'availability_zones': [],
        'description': '',
        'distributed': False,
        'external_gateway_info': None,
        'flavor_id': None,
        'ha': False,
        'id': router_id,
        'name': router_name,
        'project_id': '861808a93da0484ea1767967c4df8a23',
        'routes': [{"destination": "179.24.1.0/24", "nexthop": "172.24.3.99"}],
        'status': 'ACTIVE',
        'tenant_id': '861808a93da0484ea1767967c4df8a23',
    }

    mock_router_interface_rep = {
        'network_id': '53aee281-b06d-47fc-9e1a-37f045182b8e',
        'subnet_id': '1f1696eb-7f47-47f6-835c-4889bff88604',
        'tenant_id': '861808a93da0484ea1767967c4df8a23',
        'subnet_ids': [subnet_id],
        'port_id': '23999891-78b3-4a6b-818d-d1b713f67848',
        'id': '57076620-dcfb-42ed-8ad6-79ccb4a79ed2',
        'request_ids': ['req-f1b0b1b4-ae51-4ef9-b371-0cc3c3402cf7'],
    }

    router_availability_zone_extension = {
        "alias": "router_availability_zone",
        "updated": "2015-01-01T10:00:00-00:00",
        "description": "Availability zone support for router.",
        "links": [],
        "name": "Router Availability Zone",
    }

    router_extraroute_extension = {
        "alias": "extraroute",
        "updated": "2015-01-01T10:00:00-00:00",
        "description": "extra routes extension for router.",
        "links": [],
        "name": "Extra Routes",
    }

    enabled_neutron_extensions = [
        router_availability_zone_extension,
        router_extraroute_extension,
    ]

    def _compare_routers(self, exp, real):
        self.assertDictEqual(
            _router.Router(**exp).to_dict(computed=False),
            real.to_dict(computed=False),
        )

    def test_get_router(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers', self.router_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers'],
                        qs_elements=[f'name={self.router_name}'],
                    ),
                    json={'routers': [self.mock_router_rep]},
                ),
            ]
        )
        r = self.cloud.get_router(self.router_name)
        self.assertIsNotNone(r)
        self._compare_routers(self.mock_router_rep, r)
        self.assert_calls()

    def test_get_router_not_found(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers', 'mickey'],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers'],
                        qs_elements=['name=mickey'],
                    ),
                    json={'routers': []},
                ),
            ]
        )
        r = self.cloud.get_router('mickey')
        self.assertIsNone(r)
        self.assert_calls()

    def test_create_router(self):
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'routers']
                    ),
                    json={'router': self.mock_router_rep},
                    validate=dict(
                        json={
                            'router': {
                                'name': self.router_name,
                                'admin_state_up': True,
                            }
                        }
                    ),
                )
            ]
        )
        new_router = self.cloud.create_router(
            name=self.router_name, admin_state_up=True
        )

        self._compare_routers(self.mock_router_rep, new_router)
        self.assert_calls()

    def test_create_router_specific_tenant(self):
        new_router_tenant_id = "project_id_value"
        mock_router_rep = copy.copy(self.mock_router_rep)
        mock_router_rep['tenant_id'] = new_router_tenant_id
        mock_router_rep['project_id'] = new_router_tenant_id
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'routers']
                    ),
                    json={'router': mock_router_rep},
                    validate=dict(
                        json={
                            'router': {
                                'name': self.router_name,
                                'admin_state_up': True,
                                'project_id': new_router_tenant_id,
                            }
                        }
                    ),
                )
            ]
        )

        self.cloud.create_router(
            self.router_name, project_id=new_router_tenant_id
        )
        self.assert_calls()

    def test_create_router_with_availability_zone_hints(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'extensions']
                    ),
                    json={'extensions': self.enabled_neutron_extensions},
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'routers']
                    ),
                    json={'router': self.mock_router_rep},
                    validate=dict(
                        json={
                            'router': {
                                'name': self.router_name,
                                'admin_state_up': True,
                                'availability_zone_hints': ['nova'],
                            }
                        }
                    ),
                ),
            ]
        )
        self.cloud.create_router(
            name=self.router_name,
            admin_state_up=True,
            availability_zone_hints=['nova'],
        )
        self.assert_calls()

    def test_create_router_without_enable_snat(self):
        """Do not send enable_snat when not given."""
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'routers']
                    ),
                    json={'router': self.mock_router_rep},
                    validate=dict(
                        json={
                            'router': {
                                'name': self.router_name,
                                'admin_state_up': True,
                            }
                        }
                    ),
                )
            ]
        )
        self.cloud.create_router(name=self.router_name, admin_state_up=True)
        self.assert_calls()

    def test_create_router_with_enable_snat_True(self):
        """Send enable_snat when it is True."""
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'routers']
                    ),
                    json={'router': self.mock_router_rep},
                    validate=dict(
                        json={
                            'router': {
                                'name': self.router_name,
                                'admin_state_up': True,
                                'external_gateway_info': {'enable_snat': True},
                            }
                        }
                    ),
                )
            ]
        )
        self.cloud.create_router(
            name=self.router_name, admin_state_up=True, enable_snat=True
        )
        self.assert_calls()

    def test_create_router_with_enable_snat_False(self):
        """Send enable_snat when it is False."""
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'routers']
                    ),
                    json={'router': self.mock_router_rep},
                    validate=dict(
                        json={
                            'router': {
                                'name': self.router_name,
                                'external_gateway_info': {
                                    'enable_snat': False
                                },
                                'admin_state_up': True,
                            }
                        }
                    ),
                )
            ]
        )
        self.cloud.create_router(
            name=self.router_name, admin_state_up=True, enable_snat=False
        )
        self.assert_calls()

    def test_create_router_wrong_availability_zone_hints_type(self):
        azh_opts = "invalid"
        with testtools.ExpectedException(
            exceptions.SDKException,
            "Parameter 'availability_zone_hints' must be a list",
        ):
            self.cloud.create_router(
                name=self.router_name,
                admin_state_up=True,
                availability_zone_hints=azh_opts,
            )

    def test_add_router_interface(self):
        self.register_uris(
            [
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=[
                            'v2.0',
                            'routers',
                            self.router_id,
                            'add_router_interface',
                        ],
                    ),
                    json={'port': self.mock_router_interface_rep},
                    validate=dict(json={'subnet_id': self.subnet_id}),
                )
            ]
        )
        self.cloud.add_router_interface(
            {'id': self.router_id}, subnet_id=self.subnet_id
        )
        self.assert_calls()

    def test_remove_router_interface(self):
        self.register_uris(
            [
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=[
                            'v2.0',
                            'routers',
                            self.router_id,
                            'remove_router_interface',
                        ],
                    ),
                    json={'port': self.mock_router_interface_rep},
                    validate=dict(json={'subnet_id': self.subnet_id}),
                )
            ]
        )
        self.cloud.remove_router_interface(
            {'id': self.router_id}, subnet_id=self.subnet_id
        )
        self.assert_calls()

    def test_remove_router_interface_missing_argument(self):
        self.assertRaises(
            ValueError, self.cloud.remove_router_interface, {'id': '123'}
        )

    def test_update_router(self):
        new_router_name = "mickey"
        new_routes = []
        expected_router_rep = copy.copy(self.mock_router_rep)
        expected_router_rep['name'] = new_router_name
        expected_router_rep['routes'] = new_routes
        # validate_calls() asserts that these requests are done in order,
        # but the extensions call is only called if a non-None value is
        # passed in 'routes'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'extensions']
                    ),
                    json={'extensions': self.enabled_neutron_extensions},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers', self.router_id],
                    ),
                    json=self.mock_router_rep,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers', self.router_id],
                    ),
                    json={'router': expected_router_rep},
                    validate=dict(
                        json={
                            'router': {
                                'name': new_router_name,
                                'routes': new_routes,
                            }
                        }
                    ),
                ),
            ]
        )
        new_router = self.cloud.update_router(
            self.router_id, name=new_router_name, routes=new_routes
        )

        self._compare_routers(expected_router_rep, new_router)
        self.assert_calls()

    def test_delete_router(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers', self.router_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers'],
                        qs_elements=[f'name={self.router_name}'],
                    ),
                    json={'routers': [self.mock_router_rep]},
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers', self.router_id],
                    ),
                    json={},
                ),
            ]
        )
        self.assertTrue(self.cloud.delete_router(self.router_name))
        self.assert_calls()

    def test_delete_router_not_found(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers', self.router_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers'],
                        qs_elements=[f'name={self.router_name}'],
                    ),
                    json={'routers': []},
                ),
            ]
        )
        self.assertFalse(self.cloud.delete_router(self.router_name))
        self.assert_calls()

    def test_delete_router_multiple_found(self):
        router1 = dict(id='123', name='mickey')
        router2 = dict(id='456', name='mickey')
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers', 'mickey'],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'routers'],
                        qs_elements=['name=mickey'],
                    ),
                    json={'routers': [router1, router2]},
                ),
            ]
        )
        self.assertRaises(
            exceptions.SDKException, self.cloud.delete_router, 'mickey'
        )
        self.assert_calls()

    def _test_list_router_interfaces(
        self, router, interface_type, expected_result=None
    ):
        internal_ports = [
            {
                'id': 'internal_port_id',
                'fixed_ips': [
                    {
                        'subnet_id': 'internal_subnet_id',
                        'ip_address': "10.0.0.1",
                    }
                ],
                'device_id': self.router_id,
                'device_owner': device_owner,
            }
            for device_owner in [
                'network:router_interface',
                'network:ha_router_replicated_interface',
                'network:router_interface_distributed',
            ]
        ]

        external_ports = [
            {
                'id': 'external_port_id',
                'fixed_ips': [
                    {
                        'subnet_id': 'external_subnet_id',
                        'ip_address': "1.2.3.4",
                    }
                ],
                'device_id': self.router_id,
                'device_owner': 'network:router_gateway',
            }
        ]

        if expected_result is None:
            if interface_type == "internal":
                expected_result = internal_ports
            elif interface_type == "external":
                expected_result = external_ports
            else:
                expected_result = internal_ports + external_ports

        mock_uri = dict(
            method='GET',
            uri=self.get_mock_url(
                'network',
                'public',
                append=['v2.0', 'ports'],
                qs_elements=[f"device_id={self.router_id}"],
            ),
            json={'ports': (internal_ports + external_ports)},
        )

        self.register_uris([mock_uri])
        ret = self.cloud.list_router_interfaces(router, interface_type)
        self.assertEqual(
            [_port.Port(**i).to_dict(computed=False) for i in expected_result],
            [i.to_dict(computed=False) for i in ret],
        )
        self.assert_calls()

    router = {
        'id': router_id,
        'external_gateway_info': {
            'external_fixed_ips': [
                {'subnet_id': 'external_subnet_id', 'ip_address': '1.2.3.4'}
            ]
        },
    }

    def test_list_router_interfaces_all(self):
        self._test_list_router_interfaces(self.router, interface_type=None)

    def test_list_router_interfaces_internal(self):
        self._test_list_router_interfaces(
            self.router, interface_type="internal"
        )

    def test_list_router_interfaces_external(self):
        self._test_list_router_interfaces(
            self.router, interface_type="external"
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.shared_file_system.v2 import share_group_snapshot
from openstack.tests.unit import base

IDENTIFIER = '38152b6d-e1b5-465f-91bc-20bca4676a2a'
EXAMPLE = {
    "id": IDENTIFIER,
    "name": "snapshot_1",
    "created_at": "2021-10-24T19:36:49.555325",
    "status": "available",
    "description": "first snapshot of sg-1",
    "project_id": "7343d2f7770b4eb6a7bc33f44dcee1e0",
    "share_group_id": "fb41512f-7c49-4304-afb1-66573c7feb14",
}


class TestShareGroupSnapshot(base.TestCase):
    def test_basic(self):
        share_group_snapshots = share_group_snapshot.ShareGroupSnapshot()
        self.assertEqual(
            'share_group_snapshot', share_group_snapshots.resource_key
        )
        self.assertEqual(
            'share_group_snapshots', share_group_snapshots.resources_key
        )
        self.assertEqual(
            '/share-group-snapshots', share_group_snapshots.base_path
        )
        self.assertTrue(share_group_snapshots.allow_create)
        self.assertTrue(share_group_snapshots.allow_fetch)
        self.assertTrue(share_group_snapshots.allow_commit)
        self.assertTrue(share_group_snapshots.allow_delete)
        self.assertTrue(share_group_snapshots.allow_list)
        self.assertFalse(share_group_snapshots.allow_head)

    def test_make_share_groups(self):
        share_group_snapshots = share_group_snapshot.ShareGroupSnapshot(
            **EXAMPLE
        )
        self.assertEqual(EXAMPLE['id'], share_group_snapshots.id)
        self.assertEqual(EXAMPLE['name'], share_group_snapshots.name)
        self.assertEqual(
            EXAMPLE['created_at'], share_group_snapshots.created_at
        )
        self.assertEqual(EXAMPLE['status'], share_group_snapshots.status)
        self.assertEqual(
            EXAMPLE['description'], share_group_snapshots.description
        )
        self.assertEqual(
            EXAMPLE['project_id'], share_group_snapshots.project_id
        )
        self.assertEqual(
            EXAMPLE['share_group_id'], share_group_snapshots.share_group_id
        )


class TestShareGroupSnapshotActions(TestShareGroupSnapshot):
    def setUp(self):
        super(TestShareGroupSnapshot, self).setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.status_code = 200
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.default_microversion = '3.0'
        self.sess.post = mock.Mock(return_value=self.resp)
        self.sess._get_connection = mock.Mock(return_value=self.cloud)

    def test_reset_status(self):
        sot = share_group_snapshot.ShareGroupSnapshot(**EXAMPLE)
        self.assertIsNone(sot.reset_status(self.sess, 'available'))
        url = f'share-group-snapshots/{IDENTIFIER}/action'
        body = {"reset_status": {"status": 'available'}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_get_members(self):
        sot = share_group_snapshot.ShareGroupSnapshot(**EXAMPLE)
        sot.get_members(self.sess)
        url = f'share-group-snapshots/{IDENTIFIER}/members'
        headers = {'Accept': ''}
        self.sess.get.assert_called_with(
            url,
            headers=headers,
            microversion=self.sess.default_microversion,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# mypy: disable-error-code="method-assign"

from openstack.network.v2 import network
from openstack.network.v2 import subnet
from openstack.tests.functional.network.v2 import common


class TestSubnet(common.TestTagNeutron):
    IPV4 = 4
    CIDR = "10.100.0.0/24"
    DNS_SERVERS = ["8.8.4.4", "8.8.8.8"]
    POOL = [{"start": "10.100.0.2", "end": "10.100.0.253"}]
    ROUTES = [{"destination": "10.101.0.0/24", "nexthop": "10.100.0.254"}]
    NET_ID = None
    SUB_ID = None

    def setUp(self):
        super().setUp()
        self.NET_NAME = self.getUniqueString()
        self.SUB_NAME = self.getUniqueString()
        self.UPDATE_NAME = self.getUniqueString()
        net = self.user_cloud.network.create_network(name=self.NET_NAME)
        assert isinstance(net, network.Network)
        self.assertEqual(self.NET_NAME, net.name)
        self.NET_ID = net.id
        sub = self.user_cloud.network.create_subnet(
            name=self.SUB_NAME,
            ip_version=self.IPV4,
            network_id=self.NET_ID,
            cidr=self.CIDR,
            dns_nameservers=self.DNS_SERVERS,
            allocation_pools=self.POOL,
            host_routes=self.ROUTES,
        )
        assert isinstance(sub, subnet.Subnet)
        self.assertEqual(self.SUB_NAME, sub.name)
        self.SUB_ID = self.ID = sub.id
        self.get_command = self.user_cloud.network.get_subnet

    def tearDown(self):
        sot = self.user_cloud.network.delete_subnet(self.SUB_ID)
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_network(
            self.NET_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        super().tearDown()

    def test_find(self):
        sot = self.user_cloud.network.find_subnet(self.SUB_NAME)
        self.assertEqual(self.SUB_ID, sot.id)

    def test_get(self):
        sot = self.user_cloud.network.get_subnet(self.SUB_ID)
        self.assertEqual(self.SUB_NAME, sot.name)
        self.assertEqual(self.SUB_ID, sot.id)
        self.assertEqual(self.DNS_SERVERS, sot.dns_nameservers)
        self.assertEqual(self.CIDR, sot.cidr)
        self.assertEqual(self.POOL, sot.allocation_pools)
        self.assertEqual(self.IPV4, sot.ip_version)
        self.assertEqual(self.ROUTES, sot.host_routes)
        self.assertEqual("10.100.0.1", sot.gateway_ip)
        self.assertTrue(sot.is_dhcp_enabled)

    def test_list(self):
        names = [o.name for o in self.user_cloud.network.subnets()]
        self.assertIn(self.SUB_NAME, names)

    def test_update(self):
        sot = self.user_cloud.network.update_subnet(
            self.SUB_ID, name=self.UPDATE_NAME
        )
        self.assertEqual(self.UPDATE_NAME, sot.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.load_balancer.v2 import listener
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'admin_state_up': True,
    'allowed_cidrs': ['192.168.1.0/24'],
    'connection_limit': '2',
    'default_pool_id': uuid.uuid4(),
    'description': 'test description',
    'id': IDENTIFIER,
    'insert_headers': {"X-Forwarded-For": "true"},
    'l7policies': [{'id': uuid.uuid4()}],
    'loadbalancers': [{'id': uuid.uuid4()}],
    'name': 'test_listener',
    'project_id': uuid.uuid4(),
    'protocol': 'TEST_PROTOCOL',
    'protocol_port': 10,
    'default_tls_container_ref': (
        'http://198.51.100.10:9311/v1/containers/'
        'a570068c-d295-4780-91d4-3046a325db51'
    ),
    'sni_container_refs': [],
    'created_at': '2017-07-17T12:14:57.233772',
    'updated_at': '2017-07-17T12:16:57.233772',
    'operating_status': 'ONLINE',
    'provisioning_status': 'ACTIVE',
    'hsts_include_subdomains': True,
    'hsts_max_age': 30_000_000,
    'hsts_preload': False,
    'timeout_client_data': 50000,
    'timeout_member_connect': 5000,
    'timeout_member_data': 50000,
    'timeout_tcp_inspect': 0,
    'tls_ciphers': 'ECDHE-RSA-AES256-GCM-SHA384:ECDHE-RSA-AES128-GCM-SHA256',
    'tls_versions': ['TLSv1.1', 'TLSv1.2'],
    'alpn_protocols': ['h2', 'http/1.1', 'http/1.0'],
}

EXAMPLE_STATS = {
    'active_connections': 1,
    'bytes_in': 2,
    'bytes_out': 3,
    'request_errors': 4,
    'total_connections': 5,
}


class TestListener(base.TestCase):
    def test_basic(self):
        test_listener = listener.Listener()
        self.assertEqual('listener', test_listener.resource_key)
        self.assertEqual('listeners', test_listener.resources_key)
        self.assertEqual('/lbaas/listeners', test_listener.base_path)
        self.assertTrue(test_listener.allow_create)
        self.assertTrue(test_listener.allow_fetch)
        self.assertTrue(test_listener.allow_commit)
        self.assertTrue(test_listener.allow_delete)
        self.assertTrue(test_listener.allow_list)

    def test_make_it(self):
        test_listener = listener.Listener(**EXAMPLE)
        self.assertTrue(test_listener.is_admin_state_up)
        self.assertEqual(EXAMPLE['allowed_cidrs'], test_listener.allowed_cidrs)
        self.assertEqual(
            EXAMPLE['connection_limit'], test_listener.connection_limit
        )
        self.assertEqual(
            EXAMPLE['default_pool_id'], test_listener.default_pool_id
        )
        self.assertEqual(EXAMPLE['description'], test_listener.description)
        self.assertEqual(EXAMPLE['id'], test_listener.id)
        self.assertEqual(
            EXAMPLE['insert_headers'], test_listener.insert_headers
        )
        self.assertEqual(EXAMPLE['l7policies'], test_listener.l7_policies)
        self.assertEqual(
            EXAMPLE['loadbalancers'], test_listener.load_balancers
        )
        self.assertEqual(EXAMPLE['name'], test_listener.name)
        self.assertEqual(EXAMPLE['project_id'], test_listener.project_id)
        self.assertEqual(EXAMPLE['protocol'], test_listener.protocol)
        self.assertEqual(EXAMPLE['protocol_port'], test_listener.protocol_port)
        self.assertEqual(
            EXAMPLE['default_tls_container_ref'],
            test_listener.default_tls_container_ref,
        )
        self.assertEqual(
            EXAMPLE['sni_container_refs'], test_listener.sni_container_refs
        )
        self.assertEqual(EXAMPLE['created_at'], test_listener.created_at)
        self.assertEqual(EXAMPLE['updated_at'], test_listener.updated_at)
        self.assertTrue(test_listener.is_hsts_include_subdomains)
        self.assertEqual(EXAMPLE['hsts_max_age'], test_listener.hsts_max_age)
        self.assertFalse(test_listener.is_hsts_preload)
        self.assertEqual(
            EXAMPLE['provisioning_status'], test_listener.provisioning_status
        )
        self.assertEqual(
            EXAMPLE['operating_status'], test_listener.operating_status
        )
        self.assertEqual(
            EXAMPLE['timeout_client_data'], test_listener.timeout_client_data
        )
        self.assertEqual(
            EXAMPLE['timeout_member_connect'],
            test_listener.timeout_member_connect,
        )
        self.assertEqual(
            EXAMPLE['timeout_member_data'], test_listener.timeout_member_data
        )
        self.assertEqual(
            EXAMPLE['timeout_tcp_inspect'], test_listener.timeout_tcp_inspect
        )
        self.assertEqual(EXAMPLE['tls_ciphers'], test_listener.tls_ciphers)
        self.assertEqual(EXAMPLE['tls_versions'], test_listener.tls_versions)
        self.assertEqual(
            EXAMPLE['alpn_protocols'], test_listener.alpn_protocols
        )

        self.assertDictEqual(
            {
                'limit': 'limit',
                'marker': 'marker',
                'created_at': 'created_at',
                'updated_at': 'updated_at',
                'description': 'description',
                'name': 'name',
                'project_id': 'project_id',
                'tags': 'tags',
                'any_tags': 'tags-any',
                'not_tags': 'not-tags',
                'not_any_tags': 'not-tags-any',
                'operating_status': 'operating_status',
                'provisioning_status': 'provisioning_status',
                'is_admin_state_up': 'admin_state_up',
                'is_hsts_include_subdomains': 'hsts_include_subdomains',
                'hsts_max_age': 'hsts_max_age',
                'is_hsts_preload': 'hsts_preload',
                'allowed_cidrs': 'allowed_cidrs',
                'connection_limit': 'connection_limit',
                'default_pool_id': 'default_pool_id',
                'default_tls_container_ref': 'default_tls_container_ref',
                'sni_container_refs': 'sni_container_refs',
                'insert_headers': 'insert_headers',
                'load_balancer_id': 'load_balancer_id',
                'protocol': 'protocol',
                'protocol_port': 'protocol_port',
                'timeout_client_data': 'timeout_client_data',
                'timeout_member_connect': 'timeout_member_connect',
                'timeout_member_data': 'timeout_member_data',
                'timeout_tcp_inspect': 'timeout_tcp_inspect',
                'tls_ciphers': 'tls_ciphers',
                'tls_versions': 'tls_versions',
                'alpn_protocols': 'alpn_protocols',
            },
            test_listener._query_mapping._mapping,
        )


class TestListenerStats(base.TestCase):
    def test_basic(self):
        test_listener = listener.ListenerStats()
        self.assertEqual('stats', test_listener.resource_key)
        self.assertEqual(
            '/lbaas/listeners/%(listener_id)s/stats', test_listener.base_path
        )
        self.assertFalse(test_listener.allow_create)
        self.assertTrue(test_listener.allow_fetch)
        self.assertFalse(test_listener.allow_delete)
        self.assertFalse(test_listener.allow_list)
        self.assertFalse(test_listener.allow_commit)

    def test_make_it(self):
        test_listener = listener.ListenerStats(**EXAMPLE_STATS)
        self.assertEqual(
            EXAMPLE_STATS['active_connections'],
            test_listener.active_connections,
        )
        self.assertEqual(EXAMPLE_STATS['bytes_in'], test_listener.bytes_in)
        self.assertEqual(EXAMPLE_STATS['bytes_out'], test_listener.bytes_out)
        self.assertEqual(
            EXAMPLE_STATS['request_errors'], test_listener.request_errors
        )
        self.assertEqual(
            EXAMPLE_STATS['total_connections'], test_listener.total_connections
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from openstack.cloud import _utils
from openstack.cloud import openstackcloud
from openstack import exceptions
from openstack import utils
from openstack import warnings as os_warnings


class IdentityCloudMixin(openstackcloud._OpenStackCloudMixin):
    def _get_project_id_param_dict(self, name_or_id):
        if name_or_id:
            project = self.get_project(name_or_id)
            if not project:
                return {}
            if utils.supports_version(self.identity, '3'):
                return {'default_project_id': project['id']}
            else:
                return {'tenant_id': project['id']}
        else:
            return {}

    def _get_domain_id_param_dict(self, domain_id):
        """Get a useable domain."""

        # Keystone v3 requires domains for user and project creation. v2 does
        # not. However, keystone v2 does not allow user creation by non-admin
        # users, so we can throw an error to the user that does not need to
        # mention api versions
        if utils.supports_version(self.identity, '3'):
            if not domain_id:
                raise exceptions.SDKException(
                    "User or project creation requires an explicit domain_id "
                    "argument."
                )
            else:
                return {'domain_id': domain_id}
        else:
            return {}

    def _get_identity_params(self, domain_id=None, project=None):
        """Get the domain and project/tenant parameters if needed.

        keystone v2 and v3 are divergent enough that we need to pass or not
        pass project or tenant_id or domain or nothing in a sane manner.
        """
        ret = {}
        ret.update(self._get_domain_id_param_dict(domain_id))
        ret.update(self._get_project_id_param_dict(project))
        return ret

    def list_projects(self, domain_id=None, name_or_id=None, filters=None):
        """List projects.

        With no parameters, returns a full listing of all visible projects.

        :param domain_id: Domain ID to scope the searched projects.
        :param name_or_id: Name or ID of the project(s).
        :param filters: A dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"

        :returns: A list of identity ``Project`` objects.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        if not filters:
            filters = {}
        query = dict(**filters)
        if name_or_id:
            query['name'] = name_or_id
        if domain_id:
            query['domain_id'] = domain_id

        return list(self.identity.projects(**query))

    def search_projects(self, name_or_id=None, filters=None, domain_id=None):
        """Backwards compatibility method for search_projects

        search_projects originally had a parameter list that was name_or_id,
        filters and list had domain_id first. This method exists in this form
        to allow code written with positional parameter to still work. But
        really, use keyword arguments.

        :param name_or_id: Name or ID of the project(s).
        :param filters: dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"
        :param domain_id: Domain ID to scope the searched projects.
        :returns: A list of identity ``Project`` objects.
        """
        projects = self.list_projects(domain_id=domain_id, filters=filters)
        return _utils._filter_list(projects, name_or_id, filters)

    def get_project(self, name_or_id, filters=None, domain_id=None):
        """Get exactly one project.

        :param name_or_id: Name or unique ID of the project.
        :param filters: **DEPRECATED** A dictionary of meta data to use for
            further filtering. Elements of this dictionary may, themselves, be
            dictionaries. Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"
        :param domain_id: Domain ID to scope the retrieved project.

        :returns: An identity ``Project`` object.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        if filters is not None:
            warnings.warn(
                "the 'filters' argument is deprecated; use "
                "'search_projects' instead",
                os_warnings.RemovedInSDK60Warning,
            )
            entities = self.search_projects(
                name_or_id, filters, domain_id=domain_id
            )
            if not entities:
                return None

            if len(entities) > 1:
                raise exceptions.SDKException(
                    f"Multiple matches found for {name_or_id}",
                )

            return entities[0]

        return self.identity.find_project(
            name_or_id=name_or_id, domain_id=domain_id
        )

    def update_project(
        self,
        name_or_id,
        enabled=None,
        domain_id=None,
        **kwargs,
    ):
        """Update a project

        :param name_or_id: Name or unique ID of the project.
        :param enabled: Whether the project is enabled or not.
        :param domain_id: Domain ID to scope the retrieved project.
        :returns: An identity ``Project`` object.
        """
        project = self.identity.find_project(
            name_or_id=name_or_id,
            domain_id=domain_id,
            ignore_missing=False,
        )
        if not project:
            raise exceptions.SDKException(f"Project {name_or_id} not found.")
        if enabled is not None:
            kwargs.update({'enabled': enabled})
        project = self.identity.update_project(project, **kwargs)
        return project

    def create_project(
        self,
        name,
        domain_id,
        description=None,
        enabled=True,
        **kwargs,
    ):
        """Create a project.

        :param name:
        :param domain_id:
        :param description:
        :param enabled:
        :returns: An identity ``Project`` object.
        """
        attrs = dict(
            name=name,
            description=description,
            domain_id=domain_id,
            is_enabled=enabled,
        )
        if kwargs:
            attrs.update(kwargs)
        return self.identity.create_project(**attrs)

    def delete_project(self, name_or_id, domain_id=None):
        """Delete a project.

        :param name_or_id: Name or unique ID of the project.
        :param domain_id: Domain ID to scope the retrieved project.

        :returns: True if delete succeeded, False if the project was not found.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call
        """
        try:
            project = self.identity.find_project(
                name_or_id=name_or_id, domain_id=domain_id, ignore_missing=True
            )
            if not project:
                self.log.debug("Project %s not found for deleting", name_or_id)
                return False
            self.identity.delete_project(project)
            return True
        except exceptions.SDKException:
            self.log.exception(f"Error in deleting project {name_or_id}")
            return False

    @_utils.valid_kwargs('domain_id', 'name')
    def list_users(self, **kwargs):
        """List users.

        :param name:
        :param domain_id: Domain ID to scope the retrieved users.

        :returns: A list of identity ``User`` objects.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        return list(self.identity.users(**kwargs))

    def search_users(self, name_or_id=None, filters=None, domain_id=None):
        """Search users.

        :param name_or_id: Name or ID of the user(s).
        :param domain_id: Domain ID to scope the retrieved users.
        :param filters: dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"

        :returns: A list of identity ``User`` objects
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        kwargs = {}
        # NOTE(jdwidari) if name_or_id isn't UUID like then make use of server-
        # side filter for user name https://bit.ly/2qh0Ijk
        # especially important when using LDAP and using page to limit results
        if name_or_id and not _utils._is_uuid_like(name_or_id):
            kwargs['name'] = name_or_id
        if domain_id:
            kwargs['domain_id'] = domain_id
        users = self.list_users(**kwargs)
        return _utils._filter_list(users, name_or_id, filters)

    # TODO(stephenfin): Remove 'filters' in a future major version
    def get_user(self, name_or_id, filters=None, domain_id=None):
        """Get exactly one user.

        :param name_or_id: Name or unique ID of the user.
        :param filters: **DEPRECATED** A dictionary of meta data to use for
            further filtering. Elements of this dictionary may, themselves, be
            dictionaries. Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"
        :param domain_id: Domain ID to scope the retrieved user.

        :returns: an identity ``User`` object
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        if filters is not None:
            warnings.warn(
                "the 'filters' argument is deprecated; use "
                "'search_user' instead",
                os_warnings.RemovedInSDK60Warning,
            )
            entities = self.search_users(
                name_or_id, filters, domain_id=domain_id
            )
            if not entities:
                return None

            if len(entities) > 1:
                raise exceptions.SDKException(
                    f"Multiple matches found for {name_or_id}",
                )

        return self.identity.find_user(name_or_id, domain_id=domain_id)

    # TODO(stephenfin): Remove normalize since it doesn't do anything
    def get_user_by_id(self, user_id, normalize=None):
        """Get a user by ID.

        :param string user_id: user ID
        :returns: an identity ``User`` object
        """
        if normalize is not None:
            warnings.warn(
                "The 'normalize' field is unnecessary and will be removed in "
                "a future release.",
                os_warnings.RemovedInSDK60Warning,
            )

        return self.identity.get_user(user_id)

    @_utils.valid_kwargs(
        'name',
        'email',
        'enabled',
        'domain_id',
        'password',
        'description',
        'default_project',
    )
    def update_user(self, name_or_id, **kwargs):
        user_kwargs = {}
        if kwargs.get('domain_id'):
            user_kwargs['domain_id'] = kwargs['domain_id']
        user = self.get_user(name_or_id, **user_kwargs)

        # TODO(mordred) When this changes to REST, force interface=admin
        # in the adapter call if it's an admin force call (and figure out how
        # to make that disctinction)
        # NOTE(samueldmq): now this is a REST call and domain_id is dropped
        # if None. keystoneclient drops keys with None values.
        if 'domain_id' in kwargs and kwargs['domain_id'] is None:
            del kwargs['domain_id']
        user = self.identity.update_user(user, **kwargs)

        return user

    def create_user(
        self,
        name,
        password=None,
        email=None,
        default_project=None,
        enabled=True,
        domain_id=None,
        description=None,
    ):
        """Create a user."""
        params = self._get_identity_params(domain_id, default_project)
        params.update({'name': name, 'email': email, 'enabled': enabled})
        if password is not None:
            params['password'] = password
        if description is not None:
            params['description'] = description

        user = self.identity.create_user(**params)

        return user

    @_utils.valid_kwargs('domain_id')
    def delete_user(self, name_or_id, **kwargs):
        try:
            user = self.get_user(name_or_id, **kwargs)
            if not user:
                self.log.debug(f"User {name_or_id} not found for deleting")
                return False

            self.identity.delete_user(user)
            return True

        except exceptions.SDKException:
            self.log.exception(f"Error in deleting user {name_or_id}")
            return False

    def _get_user_and_group(self, user_name_or_id, group_name_or_id):
        user = self.get_user(user_name_or_id)
        if not user:
            raise exceptions.SDKException(f'User {user_name_or_id} not found')

        group = self.get_group(group_name_or_id)
        if not group:
            raise exceptions.SDKException(
                f'Group {group_name_or_id} not found'
            )

        return (user, group)

    def add_user_to_group(self, name_or_id, group_name_or_id):
        """Add a user to a group.

        :param name_or_id: Name or unique ID of the user.
        :param group_name_or_id: Group name or ID

        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call
        """
        user, group = self._get_user_and_group(name_or_id, group_name_or_id)

        self.identity.add_user_to_group(user, group)

    def is_user_in_group(self, name_or_id, group_name_or_id):
        """Check to see if a user is in a group.

        :param name_or_id: Name or unique ID of the user.
        :param group_name_or_id: Group name or ID

        :returns: True if user is in the group, False otherwise
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call
        """
        user, group = self._get_user_and_group(name_or_id, group_name_or_id)

        return self.identity.check_user_in_group(user, group)

    def remove_user_from_group(self, name_or_id, group_name_or_id):
        """Remove a user from a group.

        :param name_or_id: Name or unique ID of the user.
        :param group_name_or_id: Group name or ID

        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call
        """
        user, group = self._get_user_and_group(name_or_id, group_name_or_id)

        self.identity.remove_user_from_group(user, group)

    @_utils.valid_kwargs('type', 'service_type', 'description')
    def create_service(self, name, enabled=True, **kwargs):
        """Create a service.

        :param name: Service name.
        :param type: Service type. (type or service_type required.)
        :param service_type: Service type. (type or service_type required.)
        :param description: Service description (optional).
        :param enabled: Whether the service is enabled (v3 only)

        :returns: an identity ``Service`` object
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        type_ = kwargs.pop('type', None)
        service_type = kwargs.pop('service_type', None)

        # TODO(mordred) When this changes to REST, force interface=admin
        # in the adapter call
        kwargs['type'] = type_ or service_type
        kwargs['is_enabled'] = enabled
        kwargs['name'] = name

        return self.identity.create_service(**kwargs)

    @_utils.valid_kwargs(
        'name', 'enabled', 'type', 'service_type', 'description'
    )
    def update_service(self, name_or_id, **kwargs):
        # NOTE(SamYaple): Keystone v3 only accepts 'type' but shade accepts
        #                 both 'type' and 'service_type' with a preference
        #                 towards 'type'
        type_ = kwargs.pop('type', None)
        service_type = kwargs.pop('service_type', None)
        if type_ or service_type:
            kwargs['type'] = type_ or service_type

        service = self.get_service(name_or_id)
        return self.identity.update_service(service, **kwargs)

    def list_services(self):
        """List all Keystone services.

        :returns: A list of identity ``Service`` object
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        return list(self.identity.services())

    def search_services(self, name_or_id=None, filters=None):
        """Search Keystone services.

        :param name_or_id: Name or ID of the service(s).
        :param filters: dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"

        :returns: a list of identity ``Service`` objects
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        services = self.list_services()
        return _utils._filter_list(services, name_or_id, filters)

    # TODO(stephenfin): Remove 'filters' in a future major version
    def get_service(self, name_or_id, filters=None):
        """Get exactly one Keystone service.

        :param name_or_id: Name or unique ID of the service.

        :returns: an identity ``Service`` object
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call or if multiple matches are
            found.
        """
        if filters is not None:
            warnings.warn(
                "the 'filters' argument is deprecated; use "
                "'search_services' instead",
                os_warnings.RemovedInSDK60Warning,
            )
            entities = self.search_services(name_or_id, filters)
            if not entities:
                return None

            if len(entities) > 1:
                raise exceptions.SDKException(
                    f"Multiple matches found for {name_or_id}",
                )

            return entities[0]

        return self.identity.find_service(name_or_id=name_or_id)

    def delete_service(self, name_or_id):
        """Delete a Keystone service.

        :param name_or_id: Name or unique ID of the service.

        :returns: True if delete succeeded, False otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call
        """
        service = self.get_service(name_or_id=name_or_id)
        if service is None:
            self.log.debug("Service %s not found for deleting", name_or_id)
            return False

        try:
            self.identity.delete_service(service)
            return True
        except exceptions.SDKException:
            self.log.exception(
                'Failed to delete service {id}'.format(id=service['id'])
            )
            return False

    @_utils.valid_kwargs('public_url', 'internal_url', 'admin_url')
    def create_endpoint(
        self,
        service_name_or_id,
        url=None,
        interface=None,
        region=None,
        enabled=True,
        **kwargs,
    ):
        """Create a Keystone endpoint.

        :param service_name_or_id: Service name or id for this endpoint.
        :param url: URL of the endpoint
        :param interface: Interface type of the endpoint
        :param public_url: Endpoint public URL.
        :param internal_url: Endpoint internal URL.
        :param admin_url: Endpoint admin URL.
        :param region: Endpoint region.
        :param enabled: Whether the endpoint is enabled

        :returns: A list of identity ``Endpoint`` objects
        :raises: :class:`~openstack.exceptions.SDKException` if the service
            cannot be found or if something goes wrong during the OpenStack API
            call.
        """
        public_url = kwargs.pop('public_url', None)
        internal_url = kwargs.pop('internal_url', None)
        admin_url = kwargs.pop('admin_url', None)

        if (url or interface) and (public_url or internal_url or admin_url):
            raise exceptions.SDKException(
                "create_endpoint takes either url and interface OR "
                "public_url, internal_url, admin_url"
            )

        service = self.get_service(name_or_id=service_name_or_id)
        if service is None:
            raise exceptions.SDKException(
                f"service {service_name_or_id} not found"
            )

        endpoints_args = []
        if url:
            # v3 in use, v3-like arguments, one endpoint created
            endpoints_args.append(
                {
                    'url': url,
                    'interface': interface,
                    'service_id': service['id'],
                    'enabled': enabled,
                    'region_id': region,
                }
            )
        else:
            # v3 in use, v2.0-like arguments, one endpoint created for each
            # interface url provided
            endpoint_args = {
                'region_id': region,
                'enabled': enabled,
                'service_id': service['id'],
            }
            if public_url:
                endpoint_args.update(
                    {'url': public_url, 'interface': 'public'}
                )
                endpoints_args.append(endpoint_args.copy())
            if internal_url:
                endpoint_args.update(
                    {'url': internal_url, 'interface': 'internal'}
                )
                endpoints_args.append(endpoint_args.copy())
            if admin_url:
                endpoint_args.update({'url': admin_url, 'interface': 'admin'})
                endpoints_args.append(endpoint_args.copy())

        endpoints = []
        for args in endpoints_args:
            endpoints.append(self.identity.create_endpoint(**args))
        return endpoints

    @_utils.valid_kwargs(
        'enabled', 'service_name_or_id', 'url', 'interface', 'region'
    )
    def update_endpoint(self, endpoint_id, **kwargs):
        service_name_or_id = kwargs.pop('service_name_or_id', None)
        if service_name_or_id is not None:
            kwargs['service_id'] = service_name_or_id
        if 'region' in kwargs:
            kwargs['region_id'] = kwargs.pop('region')

        return self.identity.update_endpoint(endpoint_id, **kwargs)

    def list_endpoints(self):
        """List Keystone endpoints.

        :returns: A list of identity ``Endpoint`` objects
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        return list(self.identity.endpoints())

    def search_endpoints(self, id=None, filters=None):
        """List Keystone endpoints.

        :param id: ID of endpoint(s).
        :param filters: dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"

        :returns: A list of identity ``Endpoint`` objects
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        # NOTE(SamYaple): With keystone v3 we can filter directly via the
        # the keystone api, but since the return of all the endpoints even in
        # large environments is small, we can continue to filter in shade just
        # like the v2 api.
        endpoints = self.list_endpoints()
        return _utils._filter_list(endpoints, id, filters)

    # TODO(stephenfin): Remove 'filters' since it's a noop
    def get_endpoint(self, id, filters=None):
        """Get exactly one Keystone endpoint.

        :param id: ID of endpoint.
        :returns: An identity ``Endpoint`` object
        """
        if filters is not None:
            warnings.warn(
                "the 'filters' argument is deprecated; use "
                "'search_endpoints' instead",
                os_warnings.RemovedInSDK60Warning,
            )
            entities = self.search_endpoints(id, filters)
            if not entities:
                return None

            if len(entities) > 1:
                raise exceptions.SDKException(
                    f"Multiple matches found for {id}",
                )

            return entities[0]

        return self.identity.find_endpoint(name_or_id=id)

    def delete_endpoint(self, id):
        """Delete a Keystone endpoint.

        :param id: ID of the endpoint to delete.

        :returns: True if delete succeeded, False otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        endpoint = self.get_endpoint(id=id)
        if endpoint is None:
            self.log.debug("Endpoint %s not found for deleting", id)
            return False

        try:
            self.identity.delete_endpoint(id)
            return True
        except exceptions.SDKException:
            self.log.exception(f"Failed to delete endpoint {id}")
            return False

    def create_domain(self, name, description=None, enabled=True):
        """Create a domain.

        :param name: The name of the domain.
        :param description: A description of the domain.
        :param enabled: Is the domain enabled or not (default True).
        :returns: The created identity ``Endpoint`` object.
        :raises: :class:`~openstack.exceptions.SDKException` if the domain
            cannot be created.
        """
        domain_ref = {'name': name, 'enabled': enabled}
        if description is not None:
            domain_ref['description'] = description
        return self.identity.create_domain(**domain_ref)

    # TODO(stephenfin): domain_id and name_or_id are the same thing now;
    # deprecate one of them
    def update_domain(
        self,
        domain_id=None,
        name=None,
        description=None,
        enabled=None,
        name_or_id=None,
    ):
        """Update a Keystone domain

        :param domain_id:
        :param name:
        :param description:
        :param enabled:
        :param name_or_id: Name or unique ID of the domain.
        :returns: The updated identity ``Domain`` object.
        :raises: :class:`~openstack.exceptions.SDKException` if the domain
            cannot be updated
        """
        if domain_id is None:
            if name_or_id is None:
                raise exceptions.SDKException(
                    "You must pass either domain_id or name_or_id value"
                )
            dom = self.get_domain(None, name_or_id)
            if dom is None:
                raise exceptions.SDKException(
                    f"Domain {name_or_id} not found for updating"
                )
            domain_id = dom['id']

        domain_ref = {}
        domain_ref.update({'name': name} if name else {})
        domain_ref.update({'description': description} if description else {})
        domain_ref.update({'enabled': enabled} if enabled is not None else {})
        return self.identity.update_domain(domain_id, **domain_ref)

    # TODO(stephenfin): domain_id and name_or_id are the same thing now;
    # deprecate one of them
    def delete_domain(self, domain_id=None, name_or_id=None):
        """Delete a Keystone domain.

        :param domain_id: ID of the domain to delete.
        :param name_or_id: Name or unique ID of the domain.

        :returns: True if delete succeeded, False otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        try:
            if domain_id is None:
                if name_or_id is None:
                    raise exceptions.SDKException(
                        "You must pass either domain_id or name_or_id value"
                    )
                dom = self.get_domain(name_or_id=name_or_id)
                if dom is None:
                    self.log.debug(
                        "Domain %s not found for deleting", name_or_id
                    )
                    return False
                domain_id = dom['id']

            # A domain must be disabled before deleting
            self.identity.update_domain(domain_id, is_enabled=False)
            self.identity.delete_domain(domain_id, ignore_missing=False)

            return True
        except exceptions.SDKException:
            self.log.exception(f"Failed to delete domain {domain_id}")
            raise

    def list_domains(self, **filters):
        """List Keystone domains.

        :returns: A list of identity ``Domain`` objects.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        return list(self.identity.domains(**filters))

    # TODO(stephenfin): These arguments are backwards from everything else.
    def search_domains(self, filters=None, name_or_id=None):
        """Search Keystone domains.

        :param name_or_id: Name or ID of the domain(s).
        :param filters: dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"

        :returns: a list of identity ``Domain`` objects
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        if filters is None:
            filters = {}
        if name_or_id is not None:
            domains = self.list_domains()
            return _utils._filter_list(domains, name_or_id, filters)
        else:
            return self.list_domains(**filters)

    # TODO(stephenfin): domain_id and name_or_id are the same thing now;
    # deprecate one of them
    # TODO(stephenfin): Remove 'filters' in a future major version
    def get_domain(self, domain_id=None, name_or_id=None, filters=None):
        """Get exactly one Keystone domain.

        :param domain_id: ID of the domain.
        :param name_or_id: Name or unique ID of the domain.
        :param filters: **DEPRECATED** A dictionary of meta data to use for
            further filtering. Elements of this dictionary may, themselves, be
            dictionaries. Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"

        :returns: an identity ``Domain`` object
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        if filters is not None:
            warnings.warn(
                "The 'filters' argument is deprecated for removal. It is a "
                "no-op and can be safely removed.",
                os_warnings.RemovedInSDK60Warning,
            )

        if domain_id is None:
            return self.identity.find_domain(name_or_id, ignore_missing=True)
        else:
            return self.identity.get_domain(domain_id)

    @_utils.valid_kwargs('domain_id')
    def list_groups(self, **kwargs):
        """List Keystone groups.

        :param domain_id: Domain ID.

        :returns: A list of identity ``Group`` objects
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        return list(self.identity.groups(**kwargs))

    def search_groups(self, name_or_id=None, filters=None, domain_id=None):
        """Search Keystone groups.

        :param name_or_id: Name or ID of the group(s).
        :param filters: dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"
        :param domain_id: Domain ID to scope the searched groups.

        :returns: A list of identity ``Group`` objects
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        groups = self.list_groups(domain_id=domain_id)
        return _utils._filter_list(groups, name_or_id, filters)

    # TODO(stephenfin): Remove 'filters' in a future major version
    def get_group(self, name_or_id, filters=None, domain_id=None):
        """Get exactly one Keystone group.

        :param name_or_id: Name or unique ID of the group(s).
        :param domain_id: Domain ID to scope the retrieved group.

        :returns: An identity ``Group`` object
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        if filters is not None:
            warnings.warn(
                "the 'filters' argument is deprecated; use "
                "'search_projects' instead",
                os_warnings.RemovedInSDK60Warning,
            )
            entities = self.search_groups(
                name_or_id, filters, domain_id=domain_id
            )
            if not entities:
                return None

            if len(entities) > 1:
                raise exceptions.SDKException(
                    f"Multiple matches found for {name_or_id}",
                )

            return entities[0]

        return self.identity.find_group(
            name_or_id=name_or_id, domain_id=domain_id
        )

    def create_group(self, name, description, domain=None):
        """Create a group.

        :param string name: Group name.
        :param string description: Group description.
        :param string domain: Domain name or ID for the group.

        :returns: An identity ``Group`` object
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        group_ref = {'name': name}
        if description:
            group_ref['description'] = description
        if domain:
            dom = self.get_domain(domain)
            if not dom:
                raise exceptions.SDKException(
                    f"Creating group {name} failed: Invalid domain {domain}"
                )
            group_ref['domain_id'] = dom['id']

        group = self.identity.create_group(**group_ref)

        return group

    def update_group(
        self,
        name_or_id,
        name=None,
        description=None,
        **kwargs,
    ):
        """Update an existing group

        :param name_or_id: Name or unique ID of the group.
        :param name: New group name.
        :param description: New group description.

        :returns: The updated identity ``Group`` object.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        group = self.identity.find_group(
            name_or_id, ignore_missing=False, **kwargs
        )

        group_ref = {}
        if name:
            group_ref['name'] = name
        if description:
            group_ref['description'] = description

        group = self.identity.update_group(group, **group_ref)

        return group

    def delete_group(self, name_or_id):
        """Delete a group

        :param name_or_id: Name or unique ID of the group.

        :returns: True if delete succeeded, False otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        try:
            group = self.identity.find_group(name_or_id, ignore_missing=True)
            if group is None:
                self.log.debug("Group %s not found for deleting", name_or_id)
                return False

            self.identity.delete_group(group)

            return True

        except exceptions.SDKException:
            self.log.exception(f"Unable to delete group {name_or_id}")
            return False

    def list_roles(self, **kwargs):
        """List Keystone roles.

        :returns: A list of identity ``Role`` objects
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        return list(self.identity.roles(**kwargs))

    def search_roles(self, name_or_id=None, filters=None, domain_id=None):
        """Seach Keystone roles.

        :param name: Name or ID of the role(s).
        :param filters: dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {'last_name': 'Smith', 'other': {'gender': 'Female'}}

            OR

            A string containing a jmespath expression for further filtering.
            Example::

                "[?last_name==`Smith`] | [?other.gender]==`Female`]"
        :param domain_id: Domain ID to scope the searched roles.

        :returns: a list of identity ``Role`` objects
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        roles = self.list_roles(domain_id=domain_id)
        return _utils._filter_list(roles, name_or_id, filters)

    # TODO(stephenfin): Remove 'filters' in a future major version
    def get_role(self, name_or_id, filters=None, domain_id=None):
        """Get a Keystone role.

        :param name_or_id: Name or unique ID of the role.
        :param domain_id: Domain ID to scope the retrieved role.

        :returns: An identity ``Role`` object if found, else None.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        if filters is not None:
            warnings.warn(
                "the 'filters' argument is deprecated; use "
                "'search_roles' instead",
                os_warnings.RemovedInSDK60Warning,
            )
            entities = self.search_roles(
                name_or_id, filters, domain_id=domain_id
            )
            if not entities:
                return None

            if len(entities) > 1:
                raise exceptions.SDKException(
                    f"Multiple matches found for {name_or_id}",
                )

            return entities[0]

        return self.identity.find_role(
            name_or_id=name_or_id, domain_id=domain_id
        )

    def _keystone_v3_role_assignments(self, **filters):
        # NOTE(samueldmq): different parameters have different representation
        # patterns as query parameters in the call to the list role assignments
        # API. The code below handles each set of patterns separately and
        # renames the parameters names accordingly, ignoring 'effective',
        # 'include_names' and 'include_subtree' whose do not need any renaming.
        for k in ('group', 'role', 'user'):
            if k in filters:
                try:
                    filters[k + '.id'] = filters[k].id
                except AttributeError:
                    # Also this goes away in next patches
                    filters[k + '.id'] = filters[k]
                del filters[k]
        for k in ('project', 'domain'):
            if k in filters:
                try:
                    filters['scope.' + k + '.id'] = filters[k].id
                except AttributeError:
                    # NOTE(gtema): will be dropped once domains are switched to
                    # proxy
                    filters['scope.' + k + '.id'] = filters[k]
                del filters[k]
        if 'inherited_to' in filters:
            filters['scope.OS-INHERIT:inherited_to'] = filters['inherited_to']
            del filters['inherited_to']
        elif 'os_inherit_extension_inherited_to' in filters:
            warnings.warn(
                "os_inherit_extension_inherited_to is deprecated. Use "
                "inherited_to instead.",
                os_warnings.RemovedInSDK50Warning,
            )
            filters['scope.OS-INHERIT:inherited_to'] = filters[
                'os_inherit_extension_inherited_to'
            ]
            del filters['os_inherit_extension_inherited_to']

        return list(self.identity.role_assignments(**filters))

    def list_role_assignments(self, filters=None):
        """List Keystone role assignments

        :param dict filters: Dict of filter conditions. Acceptable keys are:

            * 'user' (string) - User ID to be used as query filter.
            * 'group' (string) - Group ID to be used as query filter.
            * 'project' (string) - Project ID to be used as query filter.
            * 'domain' (string) - Domain ID to be used as query filter.
            * 'system' (string) - System name to be used as query filter.
            * 'role' (string) - Role ID to be used as query filter.
            * 'inherited_to' (string) - Return inherited
              role assignments for either 'projects' or 'domains'.
            * 'os_inherit_extension_inherited_to' (string) - Deprecated; use
              'inherited_to' instead.
            * 'effective' (boolean) - Return effective role assignments.
            * 'include_subtree' (boolean) - Include subtree

              'user' and 'group' are mutually exclusive, as are 'domain' and
              'project'.

        :returns: A list of identity
            :class:`openstack.identity.v3.role_assignment.RoleAssignment`
            objects
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        # NOTE(samueldmq): although 'include_names' is a valid query parameter
        # in the keystone v3 list role assignments API, it would have NO effect
        # on shade due to normalization. It is not documented as an acceptable
        # filter in the docs above per design!

        if not filters:
            filters = {}

        # NOTE(samueldmq): the docs above say filters are *IDs*, though if
        # dict or Resource objects are passed, this still works for backwards
        # compatibility as keystoneclient allows either IDs or objects to be
        # passed in.
        # TODO(samueldmq): fix the docs above to advertise Resource objects
        # can be provided as parameters too
        for k, v in filters.items():
            if isinstance(v, dict):
                filters[k] = v['id']

        for k in ['role', 'group', 'user']:
            if k in filters:
                filters[f'{k}_id'] = filters.pop(k)

        for k in ['domain', 'project']:
            if k in filters:
                filters[f'scope_{k}_id'] = filters.pop(k)

        if 'system' in filters:
            system_scope = filters.pop('system')
            filters['scope.system'] = system_scope

        if 'os_inherit_extension_inherited_to' in filters:
            warnings.warn(
                "os_inherit_extension_inherited_to is deprecated. Use "
                "inherited_to instead.",
                os_warnings.RemovedInSDK50Warning,
            )
            filters['inherited_to'] = filters.pop(
                'os_inherit_extension_inherited_to'
            )

        return list(self.identity.role_assignments(**filters))

    @_utils.valid_kwargs('domain_id')
    def create_role(self, name, **kwargs):
        """Create a Keystone role.

        :param string name: The name of the role.
        :param domain_id: domain id (v3)
        :returns: an identity ``Role`` object
        :raises: :class:`~openstack.exceptions.SDKException` if the role cannot
            be created
        """
        kwargs['name'] = name
        return self.identity.create_role(**kwargs)

    @_utils.valid_kwargs('domain_id')
    def update_role(self, name_or_id, name, **kwargs):
        """Update a Keystone role.

        :param name_or_id: Name or unique ID of the role.
        :param string name: The new role name
        :param domain_id: domain id
        :returns: an identity ``Role`` object
        :raises: :class:`~openstack.exceptions.SDKException` if the role cannot
            be created
        """
        role = self.get_role(name_or_id, **kwargs)
        if role is None:
            self.log.debug("Role %s not found for updating", name_or_id)
            return False

        return self.identity.update_role(role, name=name, **kwargs)

    @_utils.valid_kwargs('domain_id')
    def delete_role(self, name_or_id, **kwargs):
        """Delete a Keystone role.

        :param name_or_id: Name or unique ID of the role.
        :param domain_id: domain id (v3)

        :returns: True if delete succeeded, False otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        role = self.get_role(name_or_id, **kwargs)
        if role is None:
            self.log.debug("Role %s not found for deleting", name_or_id)
            return False

        try:
            self.identity.delete_role(role)
            return True
        except exceptions.SDKException:
            self.log.exception(f"Unable to delete role {name_or_id}")
            raise

    def _get_grant_revoke_params(
        self,
        role,
        user=None,
        group=None,
        project=None,
        domain=None,
        system=None,
    ):
        data = {}
        search_args = {}
        if domain:
            data['domain'] = self.identity.find_domain(
                domain, ignore_missing=False
            )
            # We have domain. We should use it for further searching user,
            # group, role, project
            search_args['domain_id'] = data['domain'].id

        data['role'] = self.identity.find_role(role, ignore_missing=False)

        if user and group:
            raise exceptions.SDKException(
                'Specify either a group or a user, not both'
            )
        if user is None and group is None:
            raise exceptions.SDKException(
                'Must specify either a user or a group'
            )
        if project is None and domain is None and system is None:
            raise exceptions.SDKException(
                'Must specify either a domain, project or system'
            )

        if user:
            data['user'] = self.identity.find_user(
                user, ignore_missing=False, **search_args
            )
        if group:
            data['group'] = self.identity.find_group(
                group, ignore_missing=False, **search_args
            )
        if project:
            data['project'] = self.identity.find_project(
                project, ignore_missing=False, **search_args
            )

        return data

    def grant_role(
        self,
        name_or_id,
        user=None,
        group=None,
        project=None,
        domain=None,
        system=None,
        inherited=False,
        wait=False,
        timeout=60,
    ):
        """Grant a role to a user.

        :param string name_or_id: Name or unique ID of the role.
        :param string user: The name or id of the user.
        :param string group: The name or id of the group. (v3)
        :param string project: The name or id of the project.
        :param string domain: The id of the domain. (v3)
        :param bool system: The name of the system. (v3)
        :param bool inherited: Whether the role assignment is inherited. (v3)
        :param bool wait: Wait for role to be granted
        :param int timeout: Timeout to wait for role to be granted

         NOTE: domain is a required argument when the grant is on a project,
         user or group specified by name. In that situation, they are all
         considered to be in that domain. If different domains are in use in
         the same role grant, it is required to specify those by ID.

         NOTE: for wait and timeout, sometimes granting roles is not
         instantaneous.

        NOTE: precedence is given first to project, then domain, then system

        :returns: True if the role is assigned, otherwise False
        :raises: :class:`~openstack.exceptions.SDKException` if the role cannot
            be granted
        """
        data = self._get_grant_revoke_params(
            name_or_id,
            user=user,
            group=group,
            project=project,
            domain=domain,
            system=system,
        )

        user = data.get('user')
        group = data.get('group')
        project = data.get('project')
        domain = data.get('domain')
        role = data.get('role')

        if project:
            # Proceed with project - precedence over domain and system
            if user:
                has_role = self.identity.validate_user_has_project_role(
                    project, user, role, inherited=inherited
                )
                if has_role:
                    self.log.debug('Assignment already exists')
                    return False
                self.identity.assign_project_role_to_user(
                    project, user, role, inherited=inherited
                )
            else:
                has_role = self.identity.validate_group_has_project_role(
                    project, group, role, inherited=inherited
                )
                if has_role:
                    self.log.debug('Assignment already exists')
                    return False
                self.identity.assign_project_role_to_group(
                    project, group, role, inherited=inherited
                )
        elif domain:
            # Proceed with domain - precedence over system
            if user:
                has_role = self.identity.validate_user_has_domain_role(
                    domain, user, role, inherited=inherited
                )
                if has_role:
                    self.log.debug('Assignment already exists')
                    return False
                self.identity.assign_domain_role_to_user(
                    domain, user, role, inherited=inherited
                )
            else:
                has_role = self.identity.validate_group_has_domain_role(
                    domain, group, role, inherited=inherited
                )
                if has_role:
                    self.log.debug('Assignment already exists')
                    return False
                self.identity.assign_domain_role_to_group(
                    domain, group, role, inherited=inherited
                )
        else:
            # Proceed with system
            # System name must be 'all' due to checks performed in
            # _get_grant_revoke_params
            if user:
                has_role = self.identity.validate_user_has_system_role(
                    user, role, system
                )
                if has_role:
                    self.log.debug('Assignment already exists')
                    return False
                self.identity.assign_system_role_to_user(user, role, system)
            else:
                has_role = self.identity.validate_group_has_system_role(
                    group, role, system
                )
                if has_role:
                    self.log.debug('Assignment already exists')
                    return False
                self.identity.assign_system_role_to_group(group, role, system)
        return True

    def revoke_role(
        self,
        name_or_id,
        user=None,
        group=None,
        project=None,
        domain=None,
        system=None,
        inherited=False,
        wait=False,
        timeout=60,
    ):
        """Revoke a role from a user.

        :param string name_or_id: Name or unique ID of the role.
        :param string user: The name or id of the user.
        :param string group: The name or id of the group. (v3)
        :param string project: The name or id of the project.
        :param string domain: The id of the domain. (v3)
        :param bool system: The name of the system. (v3)
        :param bool inherited: Whether the role assignment is inherited.
        :param bool wait: Wait for role to be revoked
        :param int timeout: Timeout to wait for role to be revoked

            NOTE: for wait and timeout, sometimes revoking roles is not
            instantaneous.

            NOTE: project is required for keystone v2

        NOTE: precedence is given first to project, then domain, then system

        :returns: True if the role is revoke, otherwise False
        :raises: :class:`~openstack.exceptions.SDKException` if the role cannot
            be removed
        """
        data = self._get_grant_revoke_params(
            name_or_id,
            user=user,
            group=group,
            project=project,
            domain=domain,
            system=system,
        )

        user = data.get('user')
        group = data.get('group')
        project = data.get('project')
        domain = data.get('domain')
        role = data.get('role')

        if project:
            # Proceed with project - precedence over domain and system
            if user:
                has_role = self.identity.validate_user_has_project_role(
                    project, user, role, inherited=inherited
                )
                if not has_role:
                    self.log.debug('Assignment does not exists')
                    return False
                self.identity.unassign_project_role_from_user(
                    project, user, role, inherited=inherited
                )
            else:
                has_role = self.identity.validate_group_has_project_role(
                    project, group, role, inherited=inherited
                )
                if not has_role:
                    self.log.debug('Assignment does not exists')
                    return False
                self.identity.unassign_project_role_from_group(
                    project, group, role, inherited=inherited
                )
        elif domain:
            # Proceed with domain - precedence over system
            if user:
                has_role = self.identity.validate_user_has_domain_role(
                    domain, user, role, inherited=inherited
                )
                if not has_role:
                    self.log.debug('Assignment does not exists')
                    return False
                self.identity.unassign_domain_role_from_user(
                    domain, user, role, inherited=inherited
                )
            else:
                has_role = self.identity.validate_group_has_domain_role(
                    domain, group, role, inherited=inherited
                )
                if not has_role:
                    self.log.debug('Assignment does not exists')
                    return False
                self.identity.unassign_domain_role_from_group(
                    domain, group, role, inherited=inherited
                )
        else:
            # Proceed with system
            # System name must be 'all' due to checks performed in
            # _get_grant_revoke_params
            if user:
                has_role = self.identity.validate_user_has_system_role(
                    user, role, system
                )
                if not has_role:
                    self.log.debug('Assignment does not exist')
                    return False
                self.identity.unassign_system_role_from_user(
                    user, role, system
                )
            else:
                has_role = self.identity.validate_group_has_system_role(
                    group, role, system
                )
                if not has_role:
                    self.log.debug('Assignment does not exist')
                    return False
                self.identity.unassign_system_role_from_group(
                    group, role, system
                )
        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from testscenarios import load_tests_apply_scenarios as load_tests  # noqa

from openstack import exceptions
from openstack.orchestration.v1 import _proxy
from openstack.orchestration.v1 import resource
from openstack.orchestration.v1 import software_config as sc
from openstack.orchestration.v1 import software_deployment as sd
from openstack.orchestration.v1 import stack
from openstack.orchestration.v1 import stack_environment
from openstack.orchestration.v1 import stack_event
from openstack.orchestration.v1 import stack_files
from openstack.orchestration.v1 import stack_template
from openstack.orchestration.v1 import template
from openstack import proxy
from openstack.tests.unit import test_proxy_base


class TestOrchestrationProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)


class TestOrchestrationStack(TestOrchestrationProxy):
    def test_create_stack(self):
        self.verify_create(self.proxy.create_stack, stack.Stack)

    def test_create_stack_preview(self):
        self.verify_create(
            self.proxy.create_stack,
            stack.Stack,
            method_kwargs={"preview": True, "x": 1, "y": 2, "z": 3},
            expected_kwargs={"x": 1, "y": 2, "z": 3},
        )

    def test_find_stack(self):
        self.verify_find(
            self.proxy.find_stack,
            stack.Stack,
            expected_kwargs={'resolve_outputs': True},
        )
        # mock_method="openstack.proxy.Proxy._find"
        # test_method=self.proxy.find_stack
        # method_kwargs = {
        #     'resolve_outputs': False,
        #     'ignore_missing': False
        # }
        # method_args=["name_or_id"]
        # self._verify(
        #     mock_method, test_method,
        #     method_args=method_args,
        #     method_kwargs=method_kwargs,
        #     expected_args=[stack.Stack, "name_or_id"],
        #     expected_kwargs=method_kwargs,
        #     expected_result="result")
        #
        # method_kwargs = {
        #     'resolve_outputs': True,
        #     'ignore_missing': True
        # }
        # self._verify(
        #     mock_method, test_method,
        #     method_args=method_args,
        #     method_kwargs=method_kwargs,
        #     expected_args=[stack.Stack, "name_or_id"],
        #     expected_kwargs=method_kwargs,
        #     expected_result="result")

    def test_stacks(self):
        self.verify_list(self.proxy.stacks, stack.Stack)

    def test_get_stack(self):
        self.verify_get(
            self.proxy.get_stack,
            stack.Stack,
            method_kwargs={'resolve_outputs': False},
            expected_kwargs={'resolve_outputs': False},
        )
        self.verify_get_overrided(
            self.proxy, stack.Stack, 'openstack.orchestration.v1.stack.Stack'
        )

    def test_update_stack(self):
        self._verify(
            'openstack.orchestration.v1.stack.Stack.commit',
            self.proxy.update_stack,
            expected_result='result',
            method_args=['stack'],
            method_kwargs={'preview': False},
            expected_args=[self.proxy, False],
        )

    def test_update_stack_preview(self):
        self._verify(
            'openstack.orchestration.v1.stack.Stack.commit',
            self.proxy.update_stack,
            expected_result='result',
            method_args=['stack'],
            method_kwargs={'preview': True},
            expected_args=[self.proxy, True],
        )

    def test_abandon_stack(self):
        self._verify(
            'openstack.orchestration.v1.stack.Stack.abandon',
            self.proxy.abandon_stack,
            expected_result='result',
            method_args=['stack'],
            expected_args=[self.proxy],
        )

    @mock.patch.object(stack.Stack, 'find')
    def test_export_stack_with_identity(self, mock_find):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)
        mock_find.return_value = stk

        self._verify(
            'openstack.orchestration.v1.stack.Stack.export',
            self.proxy.export_stack,
            method_args=['IDENTITY'],
            expected_args=[self.proxy],
        )
        mock_find.assert_called_once_with(
            mock.ANY, 'IDENTITY', ignore_missing=False
        )

    def test_export_stack_with_object(self):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)

        self._verify(
            'openstack.orchestration.v1.stack.Stack.export',
            self.proxy.export_stack,
            method_args=[stk],
            expected_args=[self.proxy],
        )

    def test_suspend_stack(self):
        self._verify(
            'openstack.orchestration.v1.stack.Stack.suspend',
            self.proxy.suspend_stack,
            method_args=['stack'],
            expected_args=[self.proxy],
        )

    def test_resume_stack(self):
        self._verify(
            'openstack.orchestration.v1.stack.Stack.resume',
            self.proxy.resume_stack,
            method_args=['stack'],
            expected_args=[self.proxy],
        )

    def test_delete_stack(self):
        self.verify_delete(self.proxy.delete_stack, stack.Stack, False)

    def test_delete_stack_ignore(self):
        self.verify_delete(self.proxy.delete_stack, stack.Stack, True)

    @mock.patch.object(stack.Stack, 'check')
    def test_check_stack_with_stack_object(self, mock_check):
        stk = stack.Stack(id='FAKE_ID')

        res = self.proxy.check_stack(stk)

        self.assertIsNone(res)
        mock_check.assert_called_once_with(self.proxy)

    @mock.patch.object(stack.Stack, 'existing')
    def test_check_stack_with_stack_ID(self, mock_stack):
        stk = mock.Mock()
        mock_stack.return_value = stk

        res = self.proxy.check_stack('FAKE_ID')

        self.assertIsNone(res)
        mock_stack.assert_called_once_with(id='FAKE_ID')
        stk.check.assert_called_once_with(self.proxy)


class TestOrchestrationStackEnvironment(TestOrchestrationProxy):
    @mock.patch.object(stack.Stack, 'find')
    def test_get_stack_environment_with_stack_identity(self, mock_find):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)
        mock_find.return_value = stk

        self._verify(
            'openstack.proxy.Proxy._get',
            self.proxy.get_stack_environment,
            method_args=['IDENTITY'],
            expected_args=[stack_environment.StackEnvironment],
            expected_kwargs={
                'requires_id': False,
                'stack_name': stack_name,
                'stack_id': stack_id,
            },
        )
        mock_find.assert_called_once_with(
            mock.ANY, 'IDENTITY', ignore_missing=False
        )

    def test_get_stack_environment_with_stack_object(self):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)

        self._verify(
            'openstack.proxy.Proxy._get',
            self.proxy.get_stack_environment,
            method_args=[stk],
            expected_args=[stack_environment.StackEnvironment],
            expected_kwargs={
                'requires_id': False,
                'stack_name': stack_name,
                'stack_id': stack_id,
            },
        )


class TestOrchestrationStackFiles(TestOrchestrationProxy):
    @mock.patch.object(stack_files.StackFiles, 'fetch')
    @mock.patch.object(stack.Stack, 'find')
    def test_get_stack_files_with_stack_identity(self, mock_find, mock_fetch):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)
        mock_find.return_value = stk
        mock_fetch.return_value = {'file': 'content'}

        res = self.proxy.get_stack_files('IDENTITY')

        self.assertEqual({'file': 'content'}, res)
        mock_find.assert_called_once_with(
            mock.ANY, 'IDENTITY', ignore_missing=False
        )
        mock_fetch.assert_called_once_with(self.proxy)

    @mock.patch.object(stack_files.StackFiles, 'fetch')
    def test_get_stack_files_with_stack_object(self, mock_fetch):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)
        mock_fetch.return_value = {'file': 'content'}

        res = self.proxy.get_stack_files(stk)

        self.assertEqual({'file': 'content'}, res)
        mock_fetch.assert_called_once_with(self.proxy)


class TestOrchestrationStackTemplate(TestOrchestrationProxy):
    @mock.patch.object(stack.Stack, 'find')
    def test_get_stack_template_with_stack_identity(self, mock_find):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)
        mock_find.return_value = stk

        self._verify(
            'openstack.proxy.Proxy._get',
            self.proxy.get_stack_template,
            method_args=['IDENTITY'],
            expected_args=[stack_template.StackTemplate],
            expected_kwargs={
                'requires_id': False,
                'stack_name': stack_name,
                'stack_id': stack_id,
            },
        )
        mock_find.assert_called_once_with(
            mock.ANY, 'IDENTITY', ignore_missing=False
        )

    def test_get_stack_template_with_stack_object(self):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)

        self._verify(
            'openstack.proxy.Proxy._get',
            self.proxy.get_stack_template,
            method_args=[stk],
            expected_args=[stack_template.StackTemplate],
            expected_kwargs={
                'requires_id': False,
                'stack_name': stack_name,
                'stack_id': stack_id,
            },
        )


class TestOrchestrationResource(TestOrchestrationProxy):
    @mock.patch.object(stack.Stack, 'find')
    def test_resources_with_stack_object(self, mock_find):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)

        self.verify_list(
            self.proxy.resources,
            resource.Resource,
            method_args=[stk],
            expected_args=[],
            expected_kwargs={'stack_name': stack_name, 'stack_id': stack_id},
        )

        self.assertEqual(0, mock_find.call_count)

    @mock.patch.object(stack.Stack, 'find')
    def test_resources_with_stack_name(self, mock_find):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)
        mock_find.return_value = stk

        self.verify_list(
            self.proxy.resources,
            resource.Resource,
            method_args=[stack_id],
            expected_args=[],
            expected_kwargs={'stack_name': stack_name, 'stack_id': stack_id},
        )

        mock_find.assert_called_once_with(
            mock.ANY, stack_id, ignore_missing=False
        )

    @mock.patch.object(stack.Stack, 'find')
    @mock.patch.object(resource.Resource, 'list')
    def test_resources_stack_not_found(self, mock_list, mock_find):
        stack_name = 'test_stack'
        mock_find.side_effect = exceptions.NotFoundException(
            'No stack found for test_stack'
        )

        ex = self.assertRaises(
            exceptions.NotFoundException, self.proxy.resources, stack_name
        )
        self.assertEqual('No stack found for test_stack', str(ex))


class TestOrchestrationSoftwareConfig(TestOrchestrationProxy):
    def test_create_software_config(self):
        self.verify_create(
            self.proxy.create_software_config, sc.SoftwareConfig
        )

    def test_software_configs(self):
        self.verify_list(self.proxy.software_configs, sc.SoftwareConfig)

    def test_get_software_config(self):
        self.verify_get(self.proxy.get_software_config, sc.SoftwareConfig)

    def test_delete_software_config(self):
        self.verify_delete(
            self.proxy.delete_software_config, sc.SoftwareConfig, True
        )
        self.verify_delete(
            self.proxy.delete_software_config, sc.SoftwareConfig, False
        )


class TestOrchestrationSoftwareDeployment(TestOrchestrationProxy):
    def test_create_software_deployment(self):
        self.verify_create(
            self.proxy.create_software_deployment, sd.SoftwareDeployment
        )

    def test_software_deployments(self):
        self.verify_list(
            self.proxy.software_deployments, sd.SoftwareDeployment
        )

    def test_get_software_deployment(self):
        self.verify_get(
            self.proxy.get_software_deployment, sd.SoftwareDeployment
        )

    def test_update_software_deployment(self):
        self.verify_update(
            self.proxy.update_software_deployment, sd.SoftwareDeployment
        )

    def test_delete_software_deployment(self):
        self.verify_delete(
            self.proxy.delete_software_deployment, sd.SoftwareDeployment, True
        )
        self.verify_delete(
            self.proxy.delete_software_deployment, sd.SoftwareDeployment, False
        )


class TestOrchestrationTemplate(TestOrchestrationProxy):
    @mock.patch.object(template.Template, 'validate')
    def test_validate_template(self, mock_validate):
        tmpl = mock.Mock()
        env = mock.Mock()
        tmpl_url = 'A_URI'
        ignore_errors = 'a_string'

        res = self.proxy.validate_template(tmpl, env, tmpl_url, ignore_errors)

        mock_validate.assert_called_once_with(
            self.proxy,
            tmpl,
            environment=env,
            template_url=tmpl_url,
            ignore_errors=ignore_errors,
        )
        self.assertEqual(mock_validate.return_value, res)

    def test_validate_template_no_env(self):
        tmpl = "openstack/tests/unit/orchestration/v1/hello_world.yaml"

        res = self.proxy.read_env_and_templates(tmpl)

        self.assertIsInstance(res, dict)
        self.assertIsInstance(res["files"], dict)

    def test_validate_template_invalid_request(self):
        err = self.assertRaises(
            exceptions.InvalidRequest,
            self.proxy.validate_template,
            None,
            template_url=None,
        )
        self.assertEqual(
            "'template_url' must be specified when template is None",
            str(err),
        )


class TestExtractName(TestOrchestrationProxy):
    scenarios = [
        ('stacks', dict(url='/stacks', parts=['stacks'])),
        ('name_id', dict(url='/stacks/name/id', parts=['stack'])),
        ('identity', dict(url='/stacks/id', parts=['stack'])),
        (
            'preview',
            dict(url='/stacks/name/preview', parts=['stack', 'preview']),
        ),
        (
            'stack_act',
            dict(url='/stacks/name/id/preview', parts=['stack', 'preview']),
        ),
        (
            'stack_subres',
            dict(
                url='/stacks/name/id/resources', parts=['stack', 'resources']
            ),
        ),
        (
            'stack_subres_id',
            dict(
                url='/stacks/name/id/resources/id', parts=['stack', 'resource']
            ),
        ),
        (
            'stack_subres_id_act',
            dict(
                url='/stacks/name/id/resources/id/action',
                parts=['stack', 'resource', 'action'],
            ),
        ),
        (
            'event',
            dict(
                url='/stacks/ignore/ignore/resources/ignore/events/id',
                parts=['stack', 'resource', 'event'],
            ),
        ),
        (
            'sd_metadata',
            dict(
                url='/software_deployments/metadata/ignore',
                parts=['software_deployment', 'metadata'],
            ),
        ),
    ]

    def test_extract_name(self):
        results = self.proxy._extract_name(self.url)
        self.assertEqual(self.parts, results)


class TestOrchestrationStackEvents(TestOrchestrationProxy):
    def test_stack_events_with_stack_object(self):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)

        self._verify(
            'openstack.proxy.Proxy._list',
            self.proxy.stack_events,
            method_args=[stk],
            expected_args=[stack_event.StackEvent],
            expected_kwargs={
                'stack_name': stack_name,
                'stack_id': stack_id,
            },
        )

    @mock.patch.object(proxy.Proxy, '_get')
    def test_stack_events_with_stack_id(self, mock_get):
        stack_id = '1234'
        stack_name = 'test_stack'
        stk = stack.Stack(id=stack_id, name=stack_name)
        mock_get.return_value = stk

        self._verify(
            'openstack.proxy.Proxy._list',
            self.proxy.stack_events,
            method_args=[stk],
            expected_args=[stack_event.StackEvent],
            expected_kwargs={
                'stack_name': stack_name,
                'stack_id': stack_id,
            },
        )

    def test_stack_events_with_resource_name(self):
        stack_id = '1234'
        stack_name = 'test_stack'
        resource_name = 'id'
        base_path = '/stacks/%(stack_name)s/%(stack_id)s/resources/%(resource_name)s/events'  # noqa: E501
        stk = stack.Stack(id=stack_id, name=stack_name)

        self._verify(
            'openstack.proxy.Proxy._list',
            self.proxy.stack_events,
            method_args=[stk, resource_name],
            expected_args=[stack_event.StackEvent],
            expected_kwargs={
                'stack_name': stack_name,
                'stack_id': stack_id,
                'resource_name': resource_name,
                'base_path': base_path,
            },
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.image.v2 import task
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'created_at': '2016-06-24T14:40:19Z',
    'id': IDENTIFIER,
    'input': {
        'image_properties': {'container_format': 'ovf', 'disk_format': 'vhd'},
        'import_from': 'http://example.com',
        'import_from_format': 'qcow2',
    },
    'message': 'message',
    'owner': 'fa6c8c1600f4444281658a23ee6da8e8',
    'result': 'some result',
    'schema': '/v2/schemas/task',
    'status': 'processing',
    'type': 'import',
    'updated_at': '2016-06-24T14:40:20Z',
}


class TestTask(base.TestCase):
    def test_basic(self):
        sot = task.Task()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('tasks', sot.resources_key)
        self.assertEqual('/tasks', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                'limit': 'limit',
                'marker': 'marker',
                'sort_dir': 'sort_dir',
                'sort_key': 'sort_key',
                'status': 'status',
                'type': 'type',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = task.Task(**EXAMPLE)
        self.assertEqual(IDENTIFIER, sot.id)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(EXAMPLE['input'], sot.input)
        self.assertEqual(EXAMPLE['message'], sot.message)
        self.assertEqual(EXAMPLE['owner'], sot.owner_id)
        self.assertEqual(EXAMPLE['result'], sot.result)
        self.assertEqual(EXAMPLE['schema'], sot.schema)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['type'], sot.type)
        self.assertEqual(EXAMPLE['updated_at'], sot.updated_at)
